import collections.abc
import io
import shutil
import zipfile
from contextlib import nullcontext
from difflib import get_close_matches
from pathlib import Path
from types import MappingProxyType
from typing import IO, Any, Dict, Mapping, Union, cast
from zipfile import ZipFile

import httpx
import numpy
from loguru import logger
from numpy.typing import NDArray
from pydantic import BaseModel, FilePath, NewPath, RootModel
from ruyaml import YAML
from typing_extensions import Unpack

from ._settings import settings
from .io import (
    BIOIMAGEIO_YAML,
    BioimageioYamlContent,
    BioimageioYamlContentView,
    BytesReader,
    FileDescr,
    HashKwargs,
    LightHttpFileDescr,
    OpenedBioimageioYaml,
    RelativeFilePath,
    YamlValue,
    extract_file_name,
    find_bioimageio_yaml_file_name,
    get_reader,
    identify_bioimageio_yaml_file_name,
    interprete_file_source,
)
from .io_basics import AbsoluteDirectory, FileName, ZipPath
from .types import FileSource, PermissiveFileSource
from .url import HttpUrl, RootHttpUrl
from .utils import cache
from .validation_context import ValidationContext, get_validation_context

_yaml_load = YAML(typ="safe")

_yaml_dump = YAML()
_yaml_dump.version = (1, 2)  # pyright: ignore[reportAttributeAccessIssue]
_yaml_dump.default_flow_style = False
_yaml_dump.indent(mapping=2, sequence=4, offset=2)
_yaml_dump.width = 88  # pyright: ignore[reportAttributeAccessIssue]


def read_yaml(
    file: Union[FilePath, ZipPath, IO[str], IO[bytes], BytesReader, str],
) -> YamlValue:
    if isinstance(file, (ZipPath, Path)):
        data = file.read_text(encoding="utf-8")
    else:
        data = file

    content: YamlValue = _yaml_load.load(data)
    return content


def write_yaml(
    content: Union[YamlValue, BioimageioYamlContentView, BaseModel],
    /,
    file: Union[NewPath, FilePath, IO[str], IO[bytes], ZipPath],
):
    if isinstance(file, Path):
        cm = file.open("w", encoding="utf-8")
    else:
        cm = nullcontext(file)

    if isinstance(content, BaseModel):
        content = content.model_dump(mode="json")

    with cm as f:
        _yaml_dump.dump(content, f)


def _sanitize_bioimageio_yaml(content: YamlValue) -> BioimageioYamlContent:
    if not isinstance(content, dict):
        raise ValueError(
            f"Expected {BIOIMAGEIO_YAML} content to be a mapping (got {type(content)})."
        )

    for key in content:
        if not isinstance(key, str):
            raise ValueError(
                f"Expected all keys (field names) in a {BIOIMAGEIO_YAML} "
                + f"to be strings (got '{key}' of type {type(key)})."
            )

    return cast(BioimageioYamlContent, content)


def _open_bioimageio_rdf_in_zip(
    path: ZipPath,
    *,
    original_root: Union[AbsoluteDirectory, RootHttpUrl, ZipFile],
    original_source_name: str,
) -> OpenedBioimageioYaml:
    with path.open("rb") as f:
        assert not isinstance(f, io.TextIOWrapper)
        unparsed_content = f.read().decode(encoding="utf-8")

    content = _sanitize_bioimageio_yaml(read_yaml(io.StringIO(unparsed_content)))

    return OpenedBioimageioYaml(
        content,
        original_root=original_root,
        original_file_name=extract_file_name(path),
        original_source_name=original_source_name,
        unparsed_content=unparsed_content,
    )


def _open_bioimageio_zip(
    source: ZipFile,
    *,
    original_source_name: str,
) -> OpenedBioimageioYaml:
    rdf_name = identify_bioimageio_yaml_file_name(
        [info.filename for info in source.filelist]
    )
    return _open_bioimageio_rdf_in_zip(
        ZipPath(source, rdf_name),
        original_root=source,
        original_source_name=original_source_name,
    )


def open_bioimageio_yaml(
    source: Union[PermissiveFileSource, ZipFile, ZipPath],
    /,
    **kwargs: Unpack[HashKwargs],
) -> OpenedBioimageioYaml:
    if (
        isinstance(source, str)
        and source.startswith("huggingface/")
        and source.count("/") >= 2
    ):
        if source.count("/") == 2:
            # huggingface/{user_or_org}/{repo_name}
            repo_id = source[len("huggingface/") :]
            branch = "main"
        else:
            # huggingface/{user_or_org}/{repo_id}/
            # huggingface/{user_or_org}/{repo_id}/version
            repo_id, version = source[len("huggingface/") :].rsplit("/", 1)
            if len(version) == 0:
                branch = "main"
            elif version[0].isdigit():
                branch = f"v{version}"
            else:
                branch = version

        source = HttpUrl(
            settings.huggingface_http_pattern.format(repo_id=repo_id, branch=branch)
        )

    if isinstance(source, RelativeFilePath):
        source = source.absolute()

    if isinstance(source, ZipFile):
        return _open_bioimageio_zip(source, original_source_name=str(source))
    elif isinstance(source, ZipPath):
        return _open_bioimageio_rdf_in_zip(
            source, original_root=source.root, original_source_name=str(source)
        )

    try:
        if isinstance(source, (Path, str)) and (source_dir := Path(source)).is_dir():
            # open bioimageio yaml from a folder
            src = source_dir / find_bioimageio_yaml_file_name(source_dir)
        else:
            src = interprete_file_source(source)

        reader = get_reader(src, **kwargs)

    except Exception:
        # check if `source` is a collection id
        if (
            not isinstance(source, str)
            or not isinstance(settings.id_map, str)
            or "/" not in settings.id_map
        ):
            raise

        if settings.collection_http_pattern:
            with ValidationContext(perform_io_checks=False):
                url = HttpUrl(
                    settings.collection_http_pattern.format(bioimageio_id=source)
                )

            try:
                r = httpx.get(url, follow_redirects=True)
                _ = r.raise_for_status()
                unparsed_content = r.content.decode(encoding="utf-8")
                content = _sanitize_bioimageio_yaml(read_yaml(unparsed_content))
            except Exception as e:
                logger.warning("Failed to get bioimageio.yaml from {}: {}", url, e)
            else:
                original_file_name = (
                    "rdf.yaml" if url.path is None else url.path.split("/")[-1]
                )
                return OpenedBioimageioYaml(
                    content=content,
                    original_root=url.parent,
                    original_file_name=original_file_name,
                    original_source_name=source,
                    unparsed_content=unparsed_content,
                )

        id_map = get_id_map()
        if id_map and source not in id_map:
            close_matches = get_close_matches(source, id_map)
            if len(close_matches) == 0:
                raise

            if len(close_matches) == 1:
                did_you_mean = f" Did you mean '{close_matches[0]}'?"
            else:
                did_you_mean = f" Did you mean any of {close_matches}?"

            raise FileNotFoundError(f"'{source}' not found.{did_you_mean}")

        entry = id_map[source]
        logger.info("loading {} from {}", source, entry.source)
        reader = entry.get_reader()
        with get_validation_context().replace(perform_io_checks=False):
            src = HttpUrl(entry.source)

    if reader.is_zipfile:
        return _open_bioimageio_zip(ZipFile(reader), original_source_name=str(src))

    unparsed_content = reader.read().decode(encoding="utf-8")
    content = _sanitize_bioimageio_yaml(read_yaml(unparsed_content))

    if isinstance(src, RelativeFilePath):
        src = src.absolute()

    if isinstance(src, ZipPath):
        root = src.root
    else:
        root = src.parent

    return OpenedBioimageioYaml(
        content,
        original_root=root,
        original_source_name=str(src),
        original_file_name=extract_file_name(src),
        unparsed_content=unparsed_content,
    )


_IdMap = RootModel[Dict[str, LightHttpFileDescr]]


def _get_id_map_impl(url: str) -> Dict[str, LightHttpFileDescr]:
    if not isinstance(url, str) or "/" not in url:
        logger.opt(depth=1).error("invalid id map url: {}", url)
    try:
        id_map_raw: Any = httpx.get(
            url, timeout=settings.http_timeout, follow_redirects=True
        ).json()
    except Exception as e:
        logger.opt(depth=1).error("failed to get {}: {}", url, e)
        return {}

    id_map = _IdMap.model_validate(id_map_raw)
    return id_map.root


@cache
def get_id_map() -> Mapping[str, LightHttpFileDescr]:
    try:
        if settings.resolve_draft:
            ret = _get_id_map_impl(settings.id_map_draft)
        else:
            ret = {}

        ret.update(_get_id_map_impl(settings.id_map))

    except Exception as e:
        logger.error("failed to get resource id mapping: {}", e)
        ret = {}

    return MappingProxyType(ret)


def write_content_to_zip(
    content: Mapping[
        FileName,
        Union[
            str, FilePath, ZipPath, BioimageioYamlContentView, FileDescr, BytesReader
        ],
    ],
    zip: zipfile.ZipFile,
):
    """write strings as text, dictionaries as yaml and files to a ZipFile
    Args:
        content: dict mapping archive names to local file paths,
                 strings (for text files), or dict (for yaml files).
        zip: ZipFile
    """
    for arc_name, file in content.items():
        if isinstance(file, collections.abc.Mapping):
            buf = io.StringIO()
            write_yaml(file, buf)
            file = buf.getvalue()

        if isinstance(file, str):
            zip.writestr(arc_name, file.encode("utf-8"))
        else:
            if isinstance(file, BytesReader):
                reader = file
            else:
                reader = get_reader(file)

            if (
                isinstance(reader.original_root, ZipFile)
                and reader.original_root.filename == zip.filename
                and reader.original_file_name == arc_name
            ):
                logger.debug(
                    f"Not copying {reader.original_root}/{reader.original_file_name} to itself."
                )
                continue

            with zip.open(arc_name, "w") as dest:
                shutil.copyfileobj(reader, dest, 1024 * 8)


def write_zip(
    path: Union[FilePath, IO[bytes]],
    content: Mapping[
        FileName, Union[str, FilePath, ZipPath, BioimageioYamlContentView, BytesReader]
    ],
    *,
    compression: int,
    compression_level: int,
) -> None:
    """Write a zip archive.

    Args:
        path: output path to write to.
        content: dict mapping archive names to local file paths, strings (for text files), or dict (for yaml files).
        compression: The numeric constant of compression method.
        compression_level: Compression level to use when writing files to the archive.
                           See https://docs.python.org/3/library/zipfile.html#zipfile.ZipFile

    """
    if isinstance(path, Path):
        path.parent.mkdir(parents=True, exist_ok=True)

    with ZipFile(
        path, "w", compression=compression, compresslevel=compression_level
    ) as zip:
        write_content_to_zip(content, zip)


def load_array(source: Union[FileSource, FileDescr, ZipPath]) -> NDArray[Any]:
    """load a numpy ndarray from a .npy file"""
    reader = get_reader(source)
    if settings.allow_pickle:
        logger.warning("Loading numpy array with `allow_pickle=True`.")

    return numpy.load(reader, allow_pickle=settings.allow_pickle)


def save_array(path: Union[Path, ZipPath], array: NDArray[Any]) -> None:
    """save a numpy ndarray to a .npy file"""
    with path.open(mode="wb") as f:
        assert not isinstance(f, io.TextIOWrapper)
        return numpy.save(f, array, allow_pickle=False)



from typing import Any, ClassVar, Type

from pydantic import GetCoreSchemaHandler, GetJsonSchemaHandler, RootModel
from pydantic.json_schema import JsonSchemaValue
from pydantic_core.core_schema import (
    CoreSchema,
    no_info_after_validator_function,
)
from typing_extensions import Self


class ValidatedString(str):
    root_model: ClassVar[Type[RootModel[Any]]] = RootModel[str]
    """the pydantic root model to validate the string"""
    # TODO: should we use a TypeAdapter instead?
    # TODO: with future py version:  RootModel[Any] -> RootModel[str | "literal string type"]
    _validated: Any  # initalized in __new__

    def __new__(cls, object: object):
        _validated = cls.root_model.model_validate(object).root
        self = super().__new__(cls, _validated)
        self._validated = _validated
        return self._after_validator()

    def _after_validator(self) -> Self:
        """add validation after the `root_model`"""
        return self

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        return no_info_after_validator_function(cls, handler(str))

    @classmethod
    def __get_pydantic_json_schema__(
        cls, core_schema: CoreSchema, handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        json_schema = cls.root_model.model_json_schema(mode=handler.mode)
        json_schema["title"] = cls.__name__.strip("_")
        if cls.__doc__:
            json_schema["description"] = cls.__doc__

        return json_schema



import html
import os
import platform
import subprocess
from dataclasses import dataclass
from datetime import datetime, timezone
from io import StringIO
from itertools import chain
from pathlib import Path
from tempfile import TemporaryDirectory
from textwrap import TextWrapper
from types import MappingProxyType
from typing import (
    Any,
    Dict,
    List,
    Literal,
    Mapping,
    NamedTuple,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
)

import annotated_types
import markdown
import rich.console
import rich.markdown
import rich.traceback
from loguru import logger
from pydantic import (
    BaseModel,
    Field,
    field_serializer,
    field_validator,
    model_validator,
)
from pydantic_core.core_schema import ErrorType
from typing_extensions import Annotated, Self, assert_never

from ._internal.io import is_yaml_value
from ._internal.io_utils import write_yaml
from ._internal.type_guards import is_dict
from ._internal.validation_context import ValidationContextSummary
from ._internal.warning_levels import (
    ALERT,
    ALERT_NAME,
    ERROR,
    ERROR_NAME,
    INFO,
    INFO_NAME,
    WARNING,
    WARNING_NAME,
    WarningLevel,
    WarningSeverity,
)
from ._version import VERSION
from .conda_env import CondaEnv

CONDA_CMD = "conda.bat" if platform.system() == "Windows" else "conda"

Loc = Tuple[Union[int, str], ...]
"""location of error/warning in a nested data structure"""

WarningSeverityName = Literal["info", "warning", "alert"]
WarningLevelName = Literal[WarningSeverityName, "error"]

WARNING_SEVERITY_TO_NAME: Mapping[WarningSeverity, WarningSeverityName] = (
    MappingProxyType({INFO: INFO_NAME, WARNING: WARNING_NAME, ALERT: ALERT_NAME})
)
WARNING_LEVEL_TO_NAME: Mapping[WarningLevel, WarningLevelName] = MappingProxyType(
    {INFO: INFO_NAME, WARNING: WARNING_NAME, ALERT: ALERT_NAME, ERROR: ERROR_NAME}
)
WARNING_NAME_TO_LEVEL: Mapping[WarningLevelName, WarningLevel] = MappingProxyType(
    {v: k for k, v in WARNING_LEVEL_TO_NAME.items()}
)


class ValidationEntry(BaseModel):
    """Base of `ErrorEntry` and `WarningEntry`"""

    loc: Loc
    msg: str
    type: Union[ErrorType, str]


class ErrorEntry(ValidationEntry):
    """An error in a `ValidationDetail`"""

    with_traceback: bool = False
    traceback_md: str = ""
    traceback_html: str = ""
    # private rich traceback that is not serialized
    _traceback_rich: Optional[rich.traceback.Traceback] = None

    @property
    def traceback_rich(self) -> Optional[rich.traceback.Traceback]:
        return self._traceback_rich

    def model_post_init(self, __context: Any):
        if self.with_traceback and not (self.traceback_md or self.traceback_html):
            self._traceback_rich = rich.traceback.Traceback()
            console = rich.console.Console(
                record=True,
                file=open(os.devnull, "wt", encoding="utf-8"),
                color_system="truecolor",
                width=120,
                tab_size=4,
                soft_wrap=True,
            )
            console.print(self._traceback_rich)
            if not self.traceback_md:
                self.traceback_md = console.export_text(clear=False)

            if not self.traceback_html:
                self.traceback_html = console.export_html(clear=False)


class WarningEntry(ValidationEntry):
    """A warning in a `ValidationDetail`"""

    severity: WarningSeverity = WARNING

    @property
    def severity_name(self) -> WarningSeverityName:
        return WARNING_SEVERITY_TO_NAME[self.severity]


def format_loc(
    loc: Loc, target: Union[Literal["md", "html", "plain"], rich.console.Console]
) -> str:
    """helper to format a location tuple **loc**"""
    loc_str = ".".join(f"({x})" if x[0].isupper() else x for x in map(str, loc))

    # additional field validation can make the location information quite convoluted, e.g.
    # `weights.pytorch_state_dict.dependencies.source.function-after[validate_url_ok(), url['http','https']]` Input should be a valid URL, relative URL without a base
    # therefore we remove the `.function-after[validate_url_ok(), url['http','https']]` here
    loc_str, *_ = loc_str.split(".function-after")
    if loc_str:
        if target == "md" or isinstance(target, rich.console.Console):
            start = "`"
            end = "`"
        elif target == "html":
            start = "<code>"
            end = "</code>"
        elif target == "plain":
            start = ""
            end = ""
        else:
            assert_never(target)

        return f"{start}{loc_str}{end}"
    else:
        return ""


class InstalledPackage(NamedTuple):
    name: str
    version: str
    build: str = ""
    channel: str = ""


class ValidationDetail(BaseModel, extra="allow"):
    """a detail in a validation summary"""

    name: str
    status: Literal["passed", "failed"]
    loc: Loc = ()
    """location in the RDF that this detail applies to"""
    errors: List[ErrorEntry] = Field(  # pyright: ignore[reportUnknownVariableType]
        default_factory=list
    )
    warnings: List[WarningEntry] = Field(  # pyright: ignore[reportUnknownVariableType]
        default_factory=list
    )
    context: Optional[ValidationContextSummary] = None

    recommended_env: Optional[CondaEnv] = None
    """recommended conda environemnt for this validation detail"""

    saved_conda_compare: Optional[str] = None
    """output of `conda compare <recommended env>`"""

    @field_serializer("saved_conda_compare")
    def _save_conda_compare(self, value: Optional[str]):
        return self.conda_compare

    @model_validator(mode="before")
    def _load_legacy(cls, data: Any):
        if is_dict(data):
            field_name = "conda_compare"
            if (
                field_name in data
                and (saved_field_name := f"saved_{field_name}") not in data
            ):
                data[saved_field_name] = data.pop(field_name)

        return data

    @property
    def conda_compare(self) -> Optional[str]:
        if self.recommended_env is None:
            return None

        if self.saved_conda_compare is None:
            dumped_env = self.recommended_env.model_dump(mode="json")
            if is_yaml_value(dumped_env):
                with TemporaryDirectory() as d:
                    path = Path(d) / "env.yaml"
                    with path.open("w", encoding="utf-8") as f:
                        write_yaml(dumped_env, f)

                    try:
                        compare_proc = subprocess.run(
                            [CONDA_CMD, "compare", str(path)],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.STDOUT,
                            shell=False,
                            text=True,
                        )
                    except Exception as e:
                        self.saved_conda_compare = f"Failed to run `conda compare`: {e}"
                    else:
                        self.saved_conda_compare = (
                            compare_proc.stdout
                            or f"`conda compare` exited with {compare_proc.returncode}"
                        )
            else:
                self.saved_conda_compare = (
                    "Failed to dump recommended env to valid yaml"
                )

        return self.saved_conda_compare

    @property
    def status_icon(self) -> str:
        if self.status == "passed":
            return "✔️"
        else:
            return "❌"


class ValidationSummary(BaseModel, extra="allow"):
    """Summarizes output of all bioimageio validations and tests
    for one specific `ResourceDescr` instance."""

    name: str
    """Name of the validation"""

    source_name: str
    """Source of the validated bioimageio description"""

    id: Optional[str] = None
    """ID of the resource being validated"""

    type: str
    """Type of the resource being validated"""

    format_version: str
    """Format version of the resource being validated"""

    status: Literal["passed", "valid-format", "failed"]
    """Overall status of the bioimageio validation"""

    metadata_completeness: Annotated[float, annotated_types.Interval(ge=0, le=1)] = 0.0
    """Estimate of completeness of the metadata in the resource description.

    Note: This completeness estimate may change with subsequent releases
        and should be considered bioimageio.spec version specific.
    """

    details: List[ValidationDetail]
    """List of validation details"""
    env: Set[InstalledPackage] = Field(
        default_factory=lambda: {
            InstalledPackage(
                name="bioimageio.spec",
                version=VERSION,
            )
        }
    )
    """List of selected, relevant package versions"""

    saved_conda_list: Optional[str] = None

    @field_serializer("saved_conda_list")
    def _save_conda_list(self, value: Optional[str]):
        return self.conda_list

    @property
    def conda_list(self) -> str:
        if self.saved_conda_list is None:
            try:
                p = subprocess.run(
                    [CONDA_CMD, "list"],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT,
                    shell=False,
                    text=True,
                )
            except Exception as e:
                self.saved_conda_list = f"Failed to run `conda list`: {e}"
            else:
                self.saved_conda_list = (
                    p.stdout or f"`conda list` exited with {p.returncode}"
                )

        return self.saved_conda_list

    @property
    def status_icon(self) -> str:
        if self.status == "passed":
            return "✔️"
        elif self.status == "valid-format":
            return "🟡"
        else:
            return "❌"

    @property
    def errors(self) -> List[ErrorEntry]:
        return list(chain.from_iterable(d.errors for d in self.details))

    @property
    def warnings(self) -> List[WarningEntry]:
        return list(chain.from_iterable(d.warnings for d in self.details))

    def format(
        self,
        *,
        width: Optional[int] = None,
        include_conda_list: bool = False,
    ) -> str:
        """Format summary as Markdown string (alias to `format_md`)"""
        return self.format_md(width=width, include_conda_list=include_conda_list)

    def format_md(
        self,
        *,
        width: Optional[int] = None,
        include_conda_list: bool = False,
    ) -> str:
        """Format summary as Markdown string"""
        return self._format(
            width=width, target="md", include_conda_list=include_conda_list
        )

    def format_html(
        self,
        *,
        width: Optional[int] = None,
        include_conda_list: bool = False,
    ) -> str:
        md_with_html = self._format(
            target="html", width=width, include_conda_list=include_conda_list
        )
        return markdown.markdown(
            md_with_html, extensions=["tables", "fenced_code", "nl2br"]
        )

    def display(
        self,
        *,
        width: Optional[int] = None,
        include_conda_list: bool = False,
        tab_size: int = 4,
        soft_wrap: bool = True,
    ) -> None:
        try:  # render as HTML in Jupyter notebook
            from IPython.core.getipython import get_ipython
            from IPython.display import (
                display_html,  # pyright: ignore[reportUnknownVariableType]
            )
        except ImportError:
            pass
        else:
            if get_ipython() is not None:
                _ = display_html(
                    self.format_html(
                        width=width, include_conda_list=include_conda_list
                    ),
                    raw=True,
                )
                return

        # render with rich
        _ = self._format(
            target=rich.console.Console(
                width=width,
                tab_size=tab_size,
                soft_wrap=soft_wrap,
            ),
            width=width,
            include_conda_list=include_conda_list,
        )

    def add_detail(self, detail: ValidationDetail, update_status: bool = True) -> None:
        if update_status:
            if self.status == "valid-format" and detail.status == "passed":
                # once status is 'valid-format' we can only improve to 'passed'
                self.status = "passed"
            elif self.status == "passed" and detail.status == "failed":
                # once status is 'passed' it can only degrade to 'valid-format'
                self.status = "valid-format"
            # once format is 'failed' it cannot improve

        self.details.append(detail)

    def log(
        self,
        to: Union[Literal["display"], Path, Sequence[Union[Literal["display"], Path]]],
    ) -> List[Path]:
        """Convenience method to display the validation summary in the terminal and/or
        save it to disk. See `save` for details."""
        if to == "display":
            display = True
            save_to = []
        elif isinstance(to, Path):
            display = False
            save_to = [to]
        else:
            display = "display" in to
            save_to = [p for p in to if p != "display"]

        if display:
            self.display()

        return self.save(save_to)

    def save(
        self, path: Union[Path, Sequence[Path]] = Path("{id}_summary_{now}")
    ) -> List[Path]:
        """Save the validation/test summary in JSON, Markdown or HTML format.

        Returns:
            List of file paths the summary was saved to.

        Notes:
        - Format is chosen based on the suffix: `.json`, `.md`, `.html`.
        - If **path** has no suffix it is assumed to be a direcotry to which a
          `summary.json`, `summary.md` and `summary.html` are saved to.
        """
        if isinstance(path, (str, Path)):
            path = [Path(path)]

        # folder to file paths
        file_paths: List[Path] = []
        for p in path:
            if p.suffix:
                file_paths.append(p)
            else:
                file_paths.extend(
                    [
                        p / "summary.json",
                        p / "summary.md",
                        p / "summary.html",
                    ]
                )

        now = datetime.now(timezone.utc).strftime("%Y%m%dT%H%M%SZ")
        for p in file_paths:
            p = Path(str(p).format(id=self.id or "bioimageio", now=now))
            if p.suffix == ".json":
                self.save_json(p)
            elif p.suffix == ".md":
                self.save_markdown(p)
            elif p.suffix == ".html":
                self.save_html(p)
            else:
                raise ValueError(f"Unknown summary path suffix '{p.suffix}'")

        return file_paths

    def save_json(
        self, path: Path = Path("summary.json"), *, indent: Optional[int] = 2
    ) -> None:
        """Save validation/test summary as JSON file."""
        json_str = self.model_dump_json(indent=indent)
        path.parent.mkdir(exist_ok=True, parents=True)
        _ = path.write_text(json_str, encoding="utf-8")
        logger.info("Saved summary to {}", path.absolute())

    def save_markdown(self, path: Path = Path("summary.md")) -> None:
        """Save rendered validation/test summary as Markdown file."""
        formatted = self.format_md()
        path.parent.mkdir(exist_ok=True, parents=True)
        _ = path.write_text(formatted, encoding="utf-8")
        logger.info("Saved Markdown formatted summary to {}", path.absolute())

    def save_html(self, path: Path = Path("summary.html")) -> None:
        """Save rendered validation/test summary as HTML file."""
        path.parent.mkdir(exist_ok=True, parents=True)

        html = self.format_html()
        _ = path.write_text(html, encoding="utf-8")
        logger.info("Saved HTML formatted summary to {}", path.absolute())

    @classmethod
    def load_json(cls, path: Path) -> Self:
        """Load validation/test summary from a suitable JSON file"""
        json_str = Path(path).read_text(encoding="utf-8")
        return cls.model_validate_json(json_str)

    @field_validator("env", mode="before")
    def _convert_old_env(cls, value: List[Union[List[str], Dict[str, str]]]):
        """convert old style dict values of `env` for backwards compatibility"""
        if isinstance(value, list):
            return [
                (
                    (v["name"], v["version"], v.get("build", ""), v.get("channel", ""))
                    if isinstance(v, dict) and "name" in v and "version" in v
                    else v
                )
                for v in value
            ]
        else:
            return value

    def _format(
        self,
        *,
        target: Union[rich.console.Console, Literal["html", "md"]],
        width: Optional[int],
        include_conda_list: bool,
    ) -> str:
        return _format_summary(
            self,
            target=target,
            width=width or 100,
            include_conda_list=include_conda_list,
        )


def _format_summary(
    summary: ValidationSummary,
    *,
    hide_tracebacks: bool = False,  # TODO: remove?
    hide_source: bool = False,  # TODO: remove?
    hide_env: bool = False,  # TODO: remove?
    target: Union[rich.console.Console, Literal["html", "md"]] = "md",
    include_conda_list: bool,
    width: int,
) -> str:
    parts: List[str] = []
    format_table = _format_html_table if target == "html" else _format_md_table
    details_below: Dict[str, Union[str, Tuple[str, rich.traceback.Traceback]]] = {}
    left_out_details: int = 0
    left_out_details_header = "Left out details"

    def add_part(part: str):
        parts.append(part)
        if isinstance(target, rich.console.Console):
            target.print(rich.markdown.Markdown(part))

    def add_section(header: str):
        if target == "md" or isinstance(target, rich.console.Console):
            add_part(f"\n### {header}\n")
        elif target == "html":
            parts.append(f'<h3 id="{header_to_tag(header)}">{header}</h3>')
        else:
            assert_never(target)

    def header_to_tag(header: str):
        return (
            header.replace("`", "")
            .replace("(", "")
            .replace(")", "")
            .replace(" ", "-")
            .lower()
        )

    def add_as_details_below(
        title: str, text: Union[str, Tuple[str, rich.traceback.Traceback]]
    ):
        """returns a header and its tag to link to details below"""

        def make_link(header: str):
            tag = header_to_tag(header)
            if target == "md":
                return f"[{header}](#{tag})"
            elif target == "html":
                return f'<a href="#{tag}">{header}</a>'
            elif isinstance(target, rich.console.Console):
                return f"{header} below"
            else:
                assert_never(target)

        for n in range(1, 4):
            header = f"{title} {n}"
            if header in details_below:
                if details_below[header] == text:
                    return make_link(header)
            else:
                details_below[header] = text
                return make_link(header)

        nonlocal left_out_details
        left_out_details += 1
        return make_link(left_out_details_header)

    @dataclass
    class CodeCell:
        text: str

    @dataclass
    class CodeRef:
        text: str

    def format_code(
        code: str,
        lang: str = "",
        title: str = "Details",
        cell_line_limit: int = 15,
        cell_width_limit: int = 120,
    ) -> Union[CodeRef, CodeCell]:
        if not code.strip():
            return CodeCell("")

        if target == "html":
            html_lang = f' lang="{lang}"' if lang else ""
            code = f"<pre{html_lang}>{code}</pre>"
            put_below = (
                code.count("\n") > cell_line_limit
                or max(map(len, code.split("\n"))) > cell_width_limit
            )
        else:
            put_below = True
            code = f"\n```{lang}\n{code}\n```\n"

        if put_below:
            link = add_as_details_below(title, code)
            return CodeRef(f"See {link}.")
        else:
            return CodeCell(code)

    def format_traceback(entry: ErrorEntry):
        if isinstance(target, rich.console.Console):
            if entry.traceback_rich is None:
                return format_code(entry.traceback_md, title="Traceback")
            else:
                link = add_as_details_below(
                    "Traceback", (entry.traceback_md, entry.traceback_rich)
                )
                return CodeRef(f"See {link}.")

        if target == "md":
            return format_code(entry.traceback_md, title="Traceback")
        elif target == "html":
            return format_code(entry.traceback_html, title="Traceback")
        else:
            assert_never(target)

    def format_text(text: str):
        if target == "html":
            return [f"<pre>{text}</pre>"]
        else:
            return text.split("\n")

    def get_info_table():
        info_rows = [
            [summary.status_icon, summary.name.strip(".").strip()],
            ["status", summary.status],
        ]
        if not hide_source:
            info_rows.append(["source", html.escape(summary.source_name)])

        if summary.id is not None:
            info_rows.append(["id", summary.id])

        info_rows.append(["format version", f"{summary.type} {summary.format_version}"])
        if not hide_env:
            info_rows.extend([[e.name, e.version] for e in sorted(summary.env)])

        if include_conda_list:
            info_rows.append(
                ["conda list", format_code(summary.conda_list, title="Conda List").text]
            )
        return format_table(info_rows)

    def get_details_table():
        details = [["", "Location", "Details"]]

        def append_detail(
            status: str, loc: Loc, text: str, code: Union[CodeRef, CodeCell, None]
        ):
            text_lines = format_text(text)
            status_lines = [""] * len(text_lines)
            loc_lines = [""] * len(text_lines)
            status_lines[0] = status
            loc_lines[0] = format_loc(loc, target)
            for s_line, loc_line, text_line in zip(status_lines, loc_lines, text_lines):
                details.append([s_line, loc_line, text_line])

            if code is not None:
                details.append(["", "", code.text])

        for d in summary.details:
            details.append([d.status_icon, format_loc(d.loc, target), d.name])

            for entry in d.errors:
                append_detail(
                    "❌",
                    entry.loc,
                    entry.msg,
                    None if hide_tracebacks else format_traceback(entry),
                )

            for entry in d.warnings:
                append_detail("⚠", entry.loc, entry.msg, None)

            if d.recommended_env is not None:
                rec_env = StringIO()
                json_env = d.recommended_env.model_dump(
                    mode="json", exclude_defaults=True
                )
                assert is_yaml_value(json_env)
                write_yaml(json_env, rec_env)
                append_detail(
                    "",
                    d.loc,
                    f"recommended conda environment ({d.name})",
                    format_code(
                        rec_env.getvalue(),
                        lang="yaml",
                        title="Recommended Conda Environment",
                    ),
                )

            if d.conda_compare:
                wrapped_conda_compare = "\n".join(
                    TextWrapper(width=width - 4).wrap(d.conda_compare)
                )
                append_detail(
                    "",
                    d.loc,
                    f"conda compare ({d.name})",
                    format_code(
                        wrapped_conda_compare,
                        title="Conda Environment Comparison",
                    ),
                )

        return format_table(details)

    add_part(get_info_table())
    add_part(get_details_table())

    for header, text in details_below.items():
        add_section(header)
        if isinstance(text, tuple):
            assert isinstance(target, rich.console.Console)
            text, rich_obj = text
            target.print(rich_obj)
            parts.append(f"{text}\n")
        else:
            add_part(f"{text}\n")

    if left_out_details:
        parts.append(
            f"\n{left_out_details_header}\nLeft out {left_out_details} more details for brevity.\n"
        )

    return "".join(parts)


def _format_md_table(rows: List[List[str]]) -> str:
    """format `rows` as markdown table"""
    n_cols = len(rows[0])
    assert all(len(row) == n_cols for row in rows)
    col_widths = [max(max(len(row[i]) for row in rows), 3) for i in range(n_cols)]

    # fix new lines in table cell
    rows = [[line.replace("\n", "<br>") for line in r] for r in rows]

    lines = [" | ".join(rows[0][i].center(col_widths[i]) for i in range(n_cols))]
    lines.append(" | ".join("---".center(col_widths[i]) for i in range(n_cols)))
    lines.extend(
        [
            " | ".join(row[i].ljust(col_widths[i]) for i in range(n_cols))
            for row in rows[1:]
        ]
    )
    return "\n| " + " |\n| ".join(lines) + " |\n"


def _format_html_table(rows: List[List[str]]) -> str:
    """format `rows` as HTML table"""

    def get_line(cells: List[str], cell_tag: Literal["th", "td"] = "td"):
        return (
            ["  <tr>"]
            + [f"    <{cell_tag}>{c}</{cell_tag}>" for c in cells]
            + ["  </tr>"]
        )

    table = ["<table>"] + get_line(rows[0], cell_tag="th")
    for r in rows[1:]:
        table.extend(get_line(r))

    table.append("</table>")

    return "\n".join(table)



from __future__ import annotations

import collections.abc
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Dict,
    List,
    Literal,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
    cast,
)

import numpy as np
from annotated_types import Ge, Interval, MaxLen, MinLen, MultipleOf
from numpy.typing import NDArray
from pydantic import (
    AllowInfNan,
    Discriminator,
    Field,
    RootModel,
    SerializationInfo,
    SerializerFunctionWrapHandler,
    StringConstraints,
    TypeAdapter,
    ValidationInfo,
    WrapSerializer,
    field_validator,
    model_validator,
)
from typing_extensions import Annotated, Self, assert_never, get_args

from .._internal.common_nodes import (
    KwargsNode,
    Node,
    NodeWithExplicitlySetFields,
)
from .._internal.constants import SHA256_HINT
from .._internal.field_validation import validate_unique_entries
from .._internal.field_warning import issue_warning, warn
from .._internal.io import BioimageioYamlContent, WithSuffix
from .._internal.io import FileDescr as FileDescr
from .._internal.io_basics import Sha256 as Sha256
from .._internal.io_packaging import include_in_package
from .._internal.io_utils import load_array
from .._internal.packaging_context import packaging_context_var
from .._internal.types import Datetime as Datetime
from .._internal.types import FileSource_, LowerCaseIdentifier
from .._internal.types import Identifier as Identifier
from .._internal.types import LicenseId as LicenseId
from .._internal.types import NotEmpty as NotEmpty
from .._internal.url import HttpUrl as HttpUrl
from .._internal.validated_string_with_inner_node import ValidatedStringWithInnerNode
from .._internal.validator_annotations import AfterValidator, RestrictCharacters
from .._internal.version_type import Version as Version
from .._internal.warning_levels import ALERT, INFO
from ..dataset.v0_2 import VALID_COVER_IMAGE_EXTENSIONS as VALID_COVER_IMAGE_EXTENSIONS
from ..dataset.v0_2 import DatasetDescr as DatasetDescr
from ..dataset.v0_2 import LinkedDataset as LinkedDataset
from ..generic.v0_2 import AttachmentsDescr as AttachmentsDescr
from ..generic.v0_2 import Author as Author
from ..generic.v0_2 import BadgeDescr as BadgeDescr
from ..generic.v0_2 import CiteEntry as CiteEntry
from ..generic.v0_2 import Doi as Doi
from ..generic.v0_2 import GenericModelDescrBase
from ..generic.v0_2 import LinkedResource as LinkedResource
from ..generic.v0_2 import Maintainer as Maintainer
from ..generic.v0_2 import OrcidId as OrcidId
from ..generic.v0_2 import RelativeFilePath as RelativeFilePath
from ..generic.v0_2 import ResourceId as ResourceId
from ..generic.v0_2 import Uploader as Uploader
from ._v0_4_converter import convert_from_older_format


class ModelId(ResourceId):
    pass


AxesStr = Annotated[
    str, RestrictCharacters("bitczyx"), AfterValidator(validate_unique_entries)
]
AxesInCZYX = Annotated[
    str, RestrictCharacters("czyx"), AfterValidator(validate_unique_entries)
]

PostprocessingName = Literal[
    "binarize",
    "clip",
    "scale_linear",
    "sigmoid",
    "zero_mean_unit_variance",
    "scale_range",
    "scale_mean_variance",
]
PreprocessingName = Literal[
    "binarize",
    "clip",
    "scale_linear",
    "sigmoid",
    "zero_mean_unit_variance",
    "scale_range",
]


class TensorName(LowerCaseIdentifier):
    pass


class CallableFromDepencencyNode(Node):
    _submodule_adapter: ClassVar[TypeAdapter[Identifier]] = TypeAdapter(Identifier)

    module_name: str
    """The Python module that implements **callable_name**."""

    @field_validator("module_name", mode="after")
    def _check_submodules(cls, module_name: str) -> str:
        for submod in module_name.split("."):
            _ = cls._submodule_adapter.validate_python(submod)

        return module_name

    callable_name: Identifier
    """The callable Python identifier implemented in module **module_name**."""


class CallableFromDepencency(ValidatedStringWithInnerNode[CallableFromDepencencyNode]):
    _inner_node_class = CallableFromDepencencyNode
    root_model: ClassVar[Type[RootModel[Any]]] = RootModel[
        Annotated[
            str,
            StringConstraints(strip_whitespace=True, pattern=r"^.+\..+$"),
        ]
    ]

    @classmethod
    def _get_data(cls, valid_string_data: str):
        *mods, callname = valid_string_data.split(".")
        return dict(module_name=".".join(mods), callable_name=callname)

    @property
    def module_name(self):
        """The Python module that implements **callable_name**."""
        return self._inner_node.module_name

    @property
    def callable_name(self):
        """The callable Python identifier implemented in module **module_name**."""
        return self._inner_node.callable_name


class CallableFromFileNode(Node):
    source_file: Annotated[
        Union[RelativeFilePath, HttpUrl],
        Field(union_mode="left_to_right"),
        include_in_package,
    ]
    """The Python source file that implements **callable_name**."""
    callable_name: Identifier
    """The callable Python identifier implemented in **source_file**."""


class CallableFromFile(ValidatedStringWithInnerNode[CallableFromFileNode]):
    _inner_node_class = CallableFromFileNode
    root_model: ClassVar[Type[RootModel[Any]]] = RootModel[
        Annotated[
            str,
            StringConstraints(strip_whitespace=True, pattern=r"^.+:.+$"),
        ]
    ]

    @classmethod
    def _get_data(cls, valid_string_data: str):
        *file_parts, callname = valid_string_data.split(":")
        return dict(source_file=":".join(file_parts), callable_name=callname)

    @property
    def source_file(self):
        """The Python source file that implements **callable_name**."""
        return self._inner_node.source_file

    @property
    def callable_name(self):
        """The callable Python identifier implemented in **source_file**."""
        return self._inner_node.callable_name


CustomCallable = Annotated[
    Union[CallableFromFile, CallableFromDepencency], Field(union_mode="left_to_right")
]


class DependenciesNode(Node):
    manager: Annotated[NotEmpty[str], Field(examples=["conda", "maven", "pip"])]
    """Dependency manager"""

    file: FileSource_
    """Dependency file"""


class Dependencies(ValidatedStringWithInnerNode[DependenciesNode]):
    _inner_node_class = DependenciesNode
    root_model: ClassVar[Type[RootModel[Any]]] = RootModel[
        Annotated[
            str,
            StringConstraints(strip_whitespace=True, pattern=r"^.+:.+$"),
        ]
    ]

    @classmethod
    def _get_data(cls, valid_string_data: str):
        manager, *file_parts = valid_string_data.split(":")
        return dict(manager=manager, file=":".join(file_parts))

    @property
    def manager(self):
        """Dependency manager"""
        return self._inner_node.manager

    @property
    def file(self):
        """Dependency file"""
        return self._inner_node.file


WeightsFormat = Literal[
    "keras_hdf5",
    "onnx",
    "pytorch_state_dict",
    "tensorflow_js",
    "tensorflow_saved_model_bundle",
    "torchscript",
]


class WeightsEntryDescrBase(FileDescr):
    type: ClassVar[WeightsFormat]
    weights_format_name: ClassVar[str]  # human readable

    source: FileSource_
    """The weights file."""

    attachments: Annotated[
        Union[AttachmentsDescr, None],
        warn(None, "Weights entry depends on additional attachments.", ALERT),
    ] = None
    """Attachments that are specific to this weights entry."""

    authors: Union[List[Author], None] = None
    """Authors
    Either the person(s) that have trained this model resulting in the original weights file.
        (If this is the initial weights entry, i.e. it does not have a `parent`)
    Or the person(s) who have converted the weights to this weights format.
        (If this is a child weight, i.e. it has a `parent` field)
    """

    dependencies: Annotated[
        Optional[Dependencies],
        warn(
            None,
            "Custom dependencies ({value}) specified. Avoid this whenever possible "
            + "to allow execution in a wider range of software environments.",
        ),
        Field(
            examples=[
                "conda:environment.yaml",
                "maven:./pom.xml",
                "pip:./requirements.txt",
            ]
        ),
    ] = None
    """Dependency manager and dependency file, specified as `<dependency manager>:<relative file path>`."""

    parent: Annotated[
        Optional[WeightsFormat], Field(examples=["pytorch_state_dict"])
    ] = None
    """The source weights these weights were converted from.
    For example, if a model's weights were converted from the `pytorch_state_dict` format to `torchscript`,
    The `pytorch_state_dict` weights entry has no `parent` and is the parent of the `torchscript` weights.
    All weight entries except one (the initial set of weights resulting from training the model),
    need to have this field."""

    @model_validator(mode="after")
    def check_parent_is_not_self(self) -> Self:
        if self.type == self.parent:
            raise ValueError("Weights entry can't be it's own parent.")

        return self


class KerasHdf5WeightsDescr(WeightsEntryDescrBase):
    type: ClassVar[WeightsFormat] = "keras_hdf5"
    weights_format_name: ClassVar[str] = "Keras HDF5"
    tensorflow_version: Optional[Version] = None
    """TensorFlow version used to create these weights"""

    @field_validator("tensorflow_version", mode="after")
    @classmethod
    def _tfv(cls, value: Any):
        if value is None:
            issue_warning(
                "missing. Please specify the TensorFlow version"
                + " these weights were created with.",
                value=value,
                severity=ALERT,
                field="tensorflow_version",
            )
        return value


class OnnxWeightsDescr(WeightsEntryDescrBase):
    type: ClassVar[WeightsFormat] = "onnx"
    weights_format_name: ClassVar[str] = "ONNX"
    opset_version: Optional[Annotated[int, Ge(7)]] = None
    """ONNX opset version"""

    @field_validator("opset_version", mode="after")
    @classmethod
    def _ov(cls, value: Any):
        if value is None:
            issue_warning(
                "Missing ONNX opset version (aka ONNX opset number). "
                + "Please specify the ONNX opset version these weights were created"
                + " with.",
                value=value,
                severity=ALERT,
                field="opset_version",
            )
        return value


class PytorchStateDictWeightsDescr(WeightsEntryDescrBase):
    type: ClassVar[WeightsFormat] = "pytorch_state_dict"
    weights_format_name: ClassVar[str] = "Pytorch State Dict"
    architecture: CustomCallable = Field(
        examples=["my_function.py:MyNetworkClass", "my_module.submodule.get_my_model"]
    )
    """callable returning a torch.nn.Module instance.
    Local implementation: `<relative path to file>:<identifier of implementation within the file>`.
    Implementation in a dependency: `<dependency-package>.<[dependency-module]>.<identifier>`."""

    architecture_sha256: Annotated[
        Optional[Sha256],
        Field(
            description=(
                "The SHA256 of the architecture source file, if the architecture is not"
                " defined in a module listed in `dependencies`\n"
            )
            + SHA256_HINT,
        ),
    ] = None
    """The SHA256 of the architecture source file,
    if the architecture is not defined in a module listed in `dependencies`"""

    @model_validator(mode="after")
    def check_architecture_sha256(self) -> Self:
        if isinstance(self.architecture, CallableFromFile):
            if self.architecture_sha256 is None:
                raise ValueError(
                    "Missing required `architecture_sha256` for `architecture` with"
                    + " source file."
                )
        elif self.architecture_sha256 is not None:
            raise ValueError(
                "Got `architecture_sha256` for architecture that does not have a source"
                + " file."
            )

        return self

    kwargs: Dict[str, Any] = Field(
        default_factory=cast(Callable[[], Dict[str, Any]], dict)
    )
    """key word arguments for the `architecture` callable"""

    pytorch_version: Optional[Version] = None
    """Version of the PyTorch library used.
    If `depencencies` is specified it should include pytorch and the verison has to match.
    (`dependencies` overrules `pytorch_version`)"""

    @field_validator("pytorch_version", mode="after")
    @classmethod
    def _ptv(cls, value: Any):
        if value is None:
            issue_warning(
                "missing. Please specify the PyTorch version these"
                + " PyTorch state dict weights were created with.",
                value=value,
                severity=ALERT,
                field="pytorch_version",
            )
        return value


class TorchscriptWeightsDescr(WeightsEntryDescrBase):
    type: ClassVar[WeightsFormat] = "torchscript"
    weights_format_name: ClassVar[str] = "TorchScript"
    pytorch_version: Optional[Version] = None
    """Version of the PyTorch library used."""

    @field_validator("pytorch_version", mode="after")
    @classmethod
    def _ptv(cls, value: Any):
        if value is None:
            issue_warning(
                "missing. Please specify the PyTorch version these"
                + " Torchscript weights were created with.",
                value=value,
                severity=ALERT,
                field="pytorch_version",
            )
        return value


class TensorflowJsWeightsDescr(WeightsEntryDescrBase):
    type: ClassVar[WeightsFormat] = "tensorflow_js"
    weights_format_name: ClassVar[str] = "Tensorflow.js"
    tensorflow_version: Optional[Version] = None
    """Version of the TensorFlow library used."""

    @field_validator("tensorflow_version", mode="after")
    @classmethod
    def _tfv(cls, value: Any):
        if value is None:
            issue_warning(
                "missing. Please specify the TensorFlow version"
                + " these TensorflowJs weights were created with.",
                value=value,
                severity=ALERT,
                field="tensorflow_version",
            )
        return value

    source: FileSource_
    """The multi-file weights.
    All required files/folders should be a zip archive."""


class TensorflowSavedModelBundleWeightsDescr(WeightsEntryDescrBase):
    type: ClassVar[WeightsFormat] = "tensorflow_saved_model_bundle"
    weights_format_name: ClassVar[str] = "Tensorflow Saved Model"
    tensorflow_version: Optional[Version] = None
    """Version of the TensorFlow library used."""

    @field_validator("tensorflow_version", mode="after")
    @classmethod
    def _tfv(cls, value: Any):
        if value is None:
            issue_warning(
                "missing. Please specify the TensorFlow version"
                + " these Tensorflow saved model bundle weights were created with.",
                value=value,
                severity=ALERT,
                field="tensorflow_version",
            )
        return value


class WeightsDescr(Node):
    keras_hdf5: Optional[KerasHdf5WeightsDescr] = None
    onnx: Optional[OnnxWeightsDescr] = None
    pytorch_state_dict: Optional[PytorchStateDictWeightsDescr] = None
    tensorflow_js: Optional[TensorflowJsWeightsDescr] = None
    tensorflow_saved_model_bundle: Optional[TensorflowSavedModelBundleWeightsDescr] = (
        None
    )
    torchscript: Optional[TorchscriptWeightsDescr] = None

    @model_validator(mode="after")
    def check_one_entry(self) -> Self:
        if all(
            entry is None
            for entry in [
                self.keras_hdf5,
                self.onnx,
                self.pytorch_state_dict,
                self.tensorflow_js,
                self.tensorflow_saved_model_bundle,
                self.torchscript,
            ]
        ):
            raise ValueError("Missing weights entry")

        return self

    def __getitem__(
        self,
        key: WeightsFormat,
    ):
        if key == "keras_hdf5":
            ret = self.keras_hdf5
        elif key == "onnx":
            ret = self.onnx
        elif key == "pytorch_state_dict":
            ret = self.pytorch_state_dict
        elif key == "tensorflow_js":
            ret = self.tensorflow_js
        elif key == "tensorflow_saved_model_bundle":
            ret = self.tensorflow_saved_model_bundle
        elif key == "torchscript":
            ret = self.torchscript
        else:
            raise KeyError(key)

        if ret is None:
            raise KeyError(key)

        return ret

    @property
    def available_formats(self):
        return {
            **({} if self.keras_hdf5 is None else {"keras_hdf5": self.keras_hdf5}),
            **({} if self.onnx is None else {"onnx": self.onnx}),
            **(
                {}
                if self.pytorch_state_dict is None
                else {"pytorch_state_dict": self.pytorch_state_dict}
            ),
            **(
                {}
                if self.tensorflow_js is None
                else {"tensorflow_js": self.tensorflow_js}
            ),
            **(
                {}
                if self.tensorflow_saved_model_bundle is None
                else {
                    "tensorflow_saved_model_bundle": self.tensorflow_saved_model_bundle
                }
            ),
            **({} if self.torchscript is None else {"torchscript": self.torchscript}),
        }

    @property
    def missing_formats(self):
        return {
            wf for wf in get_args(WeightsFormat) if wf not in self.available_formats
        }


class ParameterizedInputShape(Node):
    """A sequence of valid shapes given by `shape_k = min + k * step for k in {0, 1, ...}`."""

    min: NotEmpty[List[int]]
    """The minimum input shape"""

    step: NotEmpty[List[int]]
    """The minimum shape change"""

    def __len__(self) -> int:
        return len(self.min)

    @model_validator(mode="after")
    def matching_lengths(self) -> Self:
        if len(self.min) != len(self.step):
            raise ValueError("`min` and `step` required to have the same length")

        return self


class ImplicitOutputShape(Node):
    """Output tensor shape depending on an input tensor shape.
    `shape(output_tensor) = shape(input_tensor) * scale + 2 * offset`"""

    reference_tensor: TensorName
    """Name of the reference tensor."""

    scale: NotEmpty[List[Optional[float]]]
    """output_pix/input_pix for each dimension.
    'null' values indicate new dimensions, whose length is defined by 2*`offset`"""

    offset: NotEmpty[List[Union[int, Annotated[float, MultipleOf(0.5)]]]]
    """Position of origin wrt to input."""

    def __len__(self) -> int:
        return len(self.scale)

    @model_validator(mode="after")
    def matching_lengths(self) -> Self:
        if len(self.scale) != len(self.offset):
            raise ValueError(
                f"scale {self.scale} has to have same length as offset {self.offset}!"
            )
        # if we have an expanded dimension, make sure that it's offet is not zero
        for sc, off in zip(self.scale, self.offset):
            if sc is None and not off:
                raise ValueError("`offset` must not be zero if `scale` is none/zero")

        return self


class TensorDescrBase(Node):
    name: TensorName
    """Tensor name. No duplicates are allowed."""

    description: str = ""

    axes: AxesStr
    """Axes identifying characters. Same length and order as the axes in `shape`.
    | axis | description |
    | --- | --- |
    |  b  |  batch (groups multiple samples) |
    |  i  |  instance/index/element |
    |  t  |  time |
    |  c  |  channel |
    |  z  |  spatial dimension z |
    |  y  |  spatial dimension y |
    |  x  |  spatial dimension x |
    """

    data_range: Optional[
        Tuple[Annotated[float, AllowInfNan(True)], Annotated[float, AllowInfNan(True)]]
    ] = None
    """Tuple `(minimum, maximum)` specifying the allowed range of the data in this tensor.
    If not specified, the full data range that can be expressed in `data_type` is allowed."""


class BinarizeKwargs(KwargsNode):
    """key word arguments for `BinarizeDescr`"""

    threshold: float
    """The fixed threshold"""


class BinarizeDescr(NodeWithExplicitlySetFields):
    """BinarizeDescr the tensor with a fixed `BinarizeKwargs.threshold`.
    Values above the threshold will be set to one, values below the threshold to zero.
    """

    implemented_name: ClassVar[Literal["binarize"]] = "binarize"
    if TYPE_CHECKING:
        name: Literal["binarize"] = "binarize"
    else:
        name: Literal["binarize"]

    kwargs: BinarizeKwargs


class ClipKwargs(KwargsNode):
    """key word arguments for `ClipDescr`"""

    min: float
    """minimum value for clipping"""
    max: float
    """maximum value for clipping"""


class ClipDescr(NodeWithExplicitlySetFields):
    """Clip tensor values to a range.

    Set tensor values below `ClipKwargs.min` to `ClipKwargs.min`
    and above `ClipKwargs.max` to `ClipKwargs.max`.
    """

    implemented_name: ClassVar[Literal["clip"]] = "clip"
    if TYPE_CHECKING:
        name: Literal["clip"] = "clip"
    else:
        name: Literal["clip"]

    kwargs: ClipKwargs


class ScaleLinearKwargs(KwargsNode):
    """key word arguments for `ScaleLinearDescr`"""

    axes: Annotated[Optional[AxesInCZYX], Field(examples=["xy"])] = None
    """The subset of axes to scale jointly.
    For example xy to scale the two image axes for 2d data jointly."""

    gain: Union[float, List[float]] = 1.0
    """multiplicative factor"""

    offset: Union[float, List[float]] = 0.0
    """additive term"""

    @model_validator(mode="after")
    def either_gain_or_offset(self) -> Self:
        if (
            self.gain == 1.0
            or isinstance(self.gain, list)
            and all(g == 1.0 for g in self.gain)
        ) and (
            self.offset == 0.0
            or isinstance(self.offset, list)
            and all(off == 0.0 for off in self.offset)
        ):
            raise ValueError(
                "Redunt linear scaling not allowd. Set `gain` != 1.0 and/or `offset` !="
                + " 0.0."
            )

        return self


class ScaleLinearDescr(NodeWithExplicitlySetFields):
    """Fixed linear scaling."""

    implemented_name: ClassVar[Literal["scale_linear"]] = "scale_linear"
    if TYPE_CHECKING:
        name: Literal["scale_linear"] = "scale_linear"
    else:
        name: Literal["scale_linear"]

    kwargs: ScaleLinearKwargs


class SigmoidDescr(NodeWithExplicitlySetFields):
    """The logistic sigmoid funciton, a.k.a. expit function."""

    implemented_name: ClassVar[Literal["sigmoid"]] = "sigmoid"
    if TYPE_CHECKING:
        name: Literal["sigmoid"] = "sigmoid"
    else:
        name: Literal["sigmoid"]

    @property
    def kwargs(self) -> KwargsNode:
        """empty kwargs"""
        return KwargsNode()


class ZeroMeanUnitVarianceKwargs(KwargsNode):
    """key word arguments for `ZeroMeanUnitVarianceDescr`"""

    mode: Literal["fixed", "per_dataset", "per_sample"] = "fixed"
    """Mode for computing mean and variance.
    |     mode    |             description              |
    | ----------- | ------------------------------------ |
    |   fixed     | Fixed values for mean and variance   |
    | per_dataset | Compute for the entire dataset       |
    | per_sample  | Compute for each sample individually |
    """
    axes: Annotated[AxesInCZYX, Field(examples=["xy"])]
    """The subset of axes to normalize jointly.
    For example `xy` to normalize the two image axes for 2d data jointly."""

    mean: Annotated[
        Union[float, NotEmpty[List[float]], None], Field(examples=[(1.1, 2.2, 3.3)])
    ] = None
    """The mean value(s) to use for `mode: fixed`.
    For example `[1.1, 2.2, 3.3]` in the case of a 3 channel image with `axes: xy`."""
    # todo: check if means match input axes (for mode 'fixed')

    std: Annotated[
        Union[float, NotEmpty[List[float]], None], Field(examples=[(0.1, 0.2, 0.3)])
    ] = None
    """The standard deviation values to use for `mode: fixed`. Analogous to mean."""

    eps: Annotated[float, Interval(gt=0, le=0.1)] = 1e-6
    """epsilon for numeric stability: `out = (tensor - mean) / (std + eps)`."""

    @model_validator(mode="after")
    def mean_and_std_match_mode(self) -> Self:
        if self.mode == "fixed" and (self.mean is None or self.std is None):
            raise ValueError("`mean` and `std` are required for `mode: fixed`.")
        elif self.mode != "fixed" and (self.mean is not None or self.std is not None):
            raise ValueError(f"`mean` and `std` not allowed for `mode: {self.mode}`")

        return self


class ZeroMeanUnitVarianceDescr(NodeWithExplicitlySetFields):
    """Subtract mean and divide by variance."""

    implemented_name: ClassVar[Literal["zero_mean_unit_variance"]] = (
        "zero_mean_unit_variance"
    )
    if TYPE_CHECKING:
        name: Literal["zero_mean_unit_variance"] = "zero_mean_unit_variance"
    else:
        name: Literal["zero_mean_unit_variance"]

    kwargs: ZeroMeanUnitVarianceKwargs


class ScaleRangeKwargs(KwargsNode):
    """key word arguments for `ScaleRangeDescr`

    For `min_percentile`=0.0 (the default) and `max_percentile`=100 (the default)
    this processing step normalizes data to the [0, 1] intervall.
    For other percentiles the normalized values will partially be outside the [0, 1]
    intervall. Use `ScaleRange` followed by `ClipDescr` if you want to limit the
    normalized values to a range.
    """

    mode: Literal["per_dataset", "per_sample"]
    """Mode for computing percentiles.
    |     mode    |             description              |
    | ----------- | ------------------------------------ |
    | per_dataset | compute for the entire dataset       |
    | per_sample  | compute for each sample individually |
    """
    axes: Annotated[AxesInCZYX, Field(examples=["xy"])]
    """The subset of axes to normalize jointly.
    For example xy to normalize the two image axes for 2d data jointly."""

    min_percentile: Annotated[Union[int, float], Interval(ge=0, lt=100)] = 0.0
    """The lower percentile used to determine the value to align with zero."""

    max_percentile: Annotated[Union[int, float], Interval(gt=1, le=100)] = 100.0
    """The upper percentile used to determine the value to align with one.
    Has to be bigger than `min_percentile`.
    The range is 1 to 100 instead of 0 to 100 to avoid mistakenly
    accepting percentiles specified in the range 0.0 to 1.0."""

    @model_validator(mode="after")
    def min_smaller_max(self, info: ValidationInfo) -> Self:
        if self.min_percentile >= self.max_percentile:
            raise ValueError(
                f"min_percentile {self.min_percentile} >= max_percentile"
                + f" {self.max_percentile}"
            )

        return self

    eps: Annotated[float, Interval(gt=0, le=0.1)] = 1e-6
    """Epsilon for numeric stability.
    `out = (tensor - v_lower) / (v_upper - v_lower + eps)`;
    with `v_lower,v_upper` values at the respective percentiles."""

    reference_tensor: Optional[TensorName] = None
    """Tensor name to compute the percentiles from. Default: The tensor itself.
    For any tensor in `inputs` only input tensor references are allowed.
    For a tensor in `outputs` only input tensor refereences are allowed if `mode: per_dataset`"""


class ScaleRangeDescr(NodeWithExplicitlySetFields):
    """Scale with percentiles."""

    implemented_name: ClassVar[Literal["scale_range"]] = "scale_range"
    if TYPE_CHECKING:
        name: Literal["scale_range"] = "scale_range"
    else:
        name: Literal["scale_range"]

    kwargs: ScaleRangeKwargs


class ScaleMeanVarianceKwargs(KwargsNode):
    """key word arguments for `ScaleMeanVarianceDescr`"""

    mode: Literal["per_dataset", "per_sample"]
    """Mode for computing mean and variance.
    |     mode    |             description              |
    | ----------- | ------------------------------------ |
    | per_dataset | Compute for the entire dataset       |
    | per_sample  | Compute for each sample individually |
    """

    reference_tensor: TensorName
    """Name of tensor to match."""

    axes: Annotated[Optional[AxesInCZYX], Field(examples=["xy"])] = None
    """The subset of axes to scale jointly.
    For example xy to normalize the two image axes for 2d data jointly.
    Default: scale all non-batch axes jointly."""

    eps: Annotated[float, Interval(gt=0, le=0.1)] = 1e-6
    """Epsilon for numeric stability:
    "`out  = (tensor - mean) / (std + eps) * (ref_std + eps) + ref_mean."""


class ScaleMeanVarianceDescr(NodeWithExplicitlySetFields):
    """Scale the tensor s.t. its mean and variance match a reference tensor."""

    implemented_name: ClassVar[Literal["scale_mean_variance"]] = "scale_mean_variance"
    if TYPE_CHECKING:
        name: Literal["scale_mean_variance"] = "scale_mean_variance"
    else:
        name: Literal["scale_mean_variance"]

    kwargs: ScaleMeanVarianceKwargs


PreprocessingDescr = Annotated[
    Union[
        BinarizeDescr,
        ClipDescr,
        ScaleLinearDescr,
        SigmoidDescr,
        ZeroMeanUnitVarianceDescr,
        ScaleRangeDescr,
    ],
    Discriminator("name"),
]
PostprocessingDescr = Annotated[
    Union[
        BinarizeDescr,
        ClipDescr,
        ScaleLinearDescr,
        SigmoidDescr,
        ZeroMeanUnitVarianceDescr,
        ScaleRangeDescr,
        ScaleMeanVarianceDescr,
    ],
    Discriminator("name"),
]


class InputTensorDescr(TensorDescrBase):
    data_type: Literal["float32", "uint8", "uint16"]
    """For now an input tensor is expected to be given as `float32`.
    The data flow in bioimage.io models is explained
    [in this diagram.](https://docs.google.com/drawings/d/1FTw8-Rn6a6nXdkZ_SkMumtcjvur9mtIhRqLwnKqZNHM/edit)."""

    shape: Annotated[
        Union[Sequence[int], ParameterizedInputShape],
        Field(
            examples=[(1, 512, 512, 1), dict(min=(1, 64, 64, 1), step=(0, 32, 32, 0))]
        ),
    ]
    """Specification of input tensor shape."""

    preprocessing: List[PreprocessingDescr] = Field(
        default_factory=cast(  # TODO: (py>3.8) use list[PreprocessingDesr]
            Callable[[], List[PreprocessingDescr]], list
        )
    )
    """Description of how this input should be preprocessed."""

    @model_validator(mode="after")
    def zero_batch_step_and_one_batch_size(self) -> Self:
        bidx = self.axes.find("b")
        if bidx == -1:
            return self

        if isinstance(self.shape, ParameterizedInputShape):
            step = self.shape.step
            shape = self.shape.min
            if step[bidx] != 0:
                raise ValueError(
                    "Input shape step has to be zero in the batch dimension (the batch"
                    + " dimension can always be increased, but `step` should specify how"
                    + " to increase the minimal shape to find the largest single batch"
                    + " shape)"
                )
        else:
            shape = self.shape

        if shape[bidx] != 1:
            raise ValueError("Input shape has to be 1 in the batch dimension b.")

        return self

    @model_validator(mode="after")
    def validate_preprocessing_kwargs(self) -> Self:
        for p in self.preprocessing:
            kwargs_axes = p.kwargs.get("axes")
            if isinstance(kwargs_axes, str) and any(
                a not in self.axes for a in kwargs_axes
            ):
                raise ValueError("`kwargs.axes` needs to be subset of `axes`")

        return self


class OutputTensorDescr(TensorDescrBase):
    data_type: Literal[
        "float32",
        "float64",
        "uint8",
        "int8",
        "uint16",
        "int16",
        "uint32",
        "int32",
        "uint64",
        "int64",
        "bool",
    ]
    """Data type.
    The data flow in bioimage.io models is explained
    [in this diagram.](https://docs.google.com/drawings/d/1FTw8-Rn6a6nXdkZ_SkMumtcjvur9mtIhRqLwnKqZNHM/edit)."""

    shape: Union[Sequence[int], ImplicitOutputShape]
    """Output tensor shape."""

    halo: Optional[Sequence[int]] = None
    """The `halo` that should be cropped from the output tensor to avoid boundary effects.
    The `halo` is to be cropped from both sides, i.e. `shape_after_crop = shape - 2 * halo`.
    To document a `halo` that is already cropped by the model `shape.offset` has to be used instead."""

    postprocessing: List[PostprocessingDescr] = Field(
        default_factory=cast(Callable[[], List[PostprocessingDescr]], list)
    )
    """Description of how this output should be postprocessed."""

    @model_validator(mode="after")
    def matching_halo_length(self) -> Self:
        if self.halo and len(self.halo) != len(self.shape):
            raise ValueError(
                f"halo {self.halo} has to have same length as shape {self.shape}!"
            )

        return self

    @model_validator(mode="after")
    def validate_postprocessing_kwargs(self) -> Self:
        for p in self.postprocessing:
            kwargs_axes = p.kwargs.get("axes", "")
            if not isinstance(kwargs_axes, str):
                raise ValueError(f"Expected {kwargs_axes} to be a string")

            if any(a not in self.axes for a in kwargs_axes):
                raise ValueError("`kwargs.axes` needs to be subset of axes")

        return self


KnownRunMode = Literal["deepimagej"]


class RunMode(Node):
    name: Annotated[
        Union[KnownRunMode, str], warn(KnownRunMode, "Unknown run mode '{value}'.")
    ]
    """Run mode name"""

    kwargs: Dict[str, Any] = Field(
        default_factory=cast(Callable[[], Dict[str, Any]], dict)
    )
    """Run mode specific key word arguments"""


class LinkedModel(Node):
    """Reference to a bioimage.io model."""

    id: Annotated[ModelId, Field(examples=["affable-shark", "ambitious-sloth"])]
    """A valid model `id` from the bioimage.io collection."""

    version_number: Optional[int] = None
    """version number (n-th published version, not the semantic version) of linked model"""


def package_weights(
    value: Node,  # Union[v0_4.WeightsDescr, v0_5.WeightsDescr]
    handler: SerializerFunctionWrapHandler,
    info: SerializationInfo,
):
    ctxt = packaging_context_var.get()
    if ctxt is not None and ctxt.weights_priority_order is not None:
        for wf in ctxt.weights_priority_order:
            w = getattr(value, wf, None)
            if w is not None:
                break
        else:
            raise ValueError(
                "None of the weight formats in `weights_priority_order`"
                + f" ({ctxt.weights_priority_order}) is present in the given model."
            )

        assert isinstance(w, Node), type(w)
        # construct WeightsDescr with new single weight format entry
        new_w = w.model_construct(**{k: v for k, v in w if k != "parent"})
        value = value.model_construct(None, **{wf: new_w})

    return handler(
        value,
        info,  # pyright: ignore[reportArgumentType]  # taken from pydantic docs
    )


class ModelDescr(GenericModelDescrBase):
    """Specification of the fields used in a bioimage.io-compliant RDF that describes AI models with pretrained weights.

    These fields are typically stored in a YAML file which we call a model resource description file (model RDF).
    """

    implemented_format_version: ClassVar[Literal["0.4.10"]] = "0.4.10"
    if TYPE_CHECKING:
        format_version: Literal["0.4.10"] = "0.4.10"
    else:
        format_version: Literal["0.4.10"]
        """Version of the bioimage.io model description specification used.
        When creating a new model always use the latest micro/patch version described here.
        The `format_version` is important for any consumer software to understand how to parse the fields.
        """

    implemented_type: ClassVar[Literal["model"]] = "model"
    if TYPE_CHECKING:
        type: Literal["model"] = "model"
    else:
        type: Literal["model"]
        """Specialized resource type 'model'"""

    id: Optional[ModelId] = None
    """bioimage.io-wide unique resource identifier
    assigned by bioimage.io; version **un**specific."""

    authors: NotEmpty[  # pyright: ignore[reportGeneralTypeIssues]  # make mandatory
        List[Author]
    ]
    """The authors are the creators of the model RDF and the primary points of contact."""

    documentation: Annotated[
        FileSource_,
        Field(
            examples=[
                "https://raw.githubusercontent.com/bioimage-io/spec-bioimage-io/main/example_descriptions/models/unet2d_nuclei_broad/README.md",
                "README.md",
            ],
        ),
    ]
    """URL or relative path to a markdown file with additional documentation.
    The recommended documentation file name is `README.md`. An `.md` suffix is mandatory.
    The documentation should include a '[#[#]]# Validation' (sub)section
    with details on how to quantitatively validate the model on unseen data."""

    inputs: NotEmpty[List[InputTensorDescr]]
    """Describes the input tensors expected by this model."""

    license: Annotated[
        Union[LicenseId, str],
        warn(LicenseId, "Unknown license id '{value}'."),
        Field(examples=["CC0-1.0", "MIT", "BSD-2-Clause"]),
    ]
    """A [SPDX license identifier](https://spdx.org/licenses/).
    We do notsupport custom license beyond the SPDX license list, if you need that please
    [open a GitHub issue](https://github.com/bioimage-io/spec-bioimage-io/issues/new/choose
    ) to discuss your intentions with the community."""

    name: Annotated[
        str,
        MinLen(1),
        warn(MinLen(5), "Name shorter than 5 characters.", INFO),
        warn(MaxLen(64), "Name longer than 64 characters.", INFO),
    ]
    """A human-readable name of this model.
    It should be no longer than 64 characters and only contain letter, number, underscore, minus or space characters."""

    outputs: NotEmpty[List[OutputTensorDescr]]
    """Describes the output tensors."""

    @field_validator("inputs", "outputs")
    @classmethod
    def unique_tensor_descr_names(
        cls, value: Sequence[Union[InputTensorDescr, OutputTensorDescr]]
    ) -> Sequence[Union[InputTensorDescr, OutputTensorDescr]]:
        unique_names = {str(v.name) for v in value}
        if len(unique_names) != len(value):
            raise ValueError("Duplicate tensor descriptor names")

        return value

    @model_validator(mode="after")
    def unique_io_names(self) -> Self:
        unique_names = {str(ss.name) for s in (self.inputs, self.outputs) for ss in s}
        if len(unique_names) != (len(self.inputs) + len(self.outputs)):
            raise ValueError("Duplicate tensor descriptor names across inputs/outputs")

        return self

    @model_validator(mode="after")
    def minimum_shape2valid_output(self) -> Self:
        tensors_by_name: Dict[
            TensorName, Union[InputTensorDescr, OutputTensorDescr]
        ] = {t.name: t for t in self.inputs + self.outputs}

        for out in self.outputs:
            if isinstance(out.shape, ImplicitOutputShape):
                ndim_ref = len(tensors_by_name[out.shape.reference_tensor].shape)
                ndim_out_ref = len(
                    [scale for scale in out.shape.scale if scale is not None]
                )
                if ndim_ref != ndim_out_ref:
                    expanded_dim_note = (
                        " Note that expanded dimensions (`scale`: null) are not"
                        + f" counted for {out.name}'sdimensionality here."
                        if None in out.shape.scale
                        else ""
                    )
                    raise ValueError(
                        f"Referenced tensor '{out.shape.reference_tensor}' with"
                        + f" {ndim_ref} dimensions does not match output tensor"
                        + f" '{out.name}' with"
                        + f" {ndim_out_ref} dimensions.{expanded_dim_note}"
                    )

            min_out_shape = self._get_min_shape(out, tensors_by_name)
            if out.halo:
                halo = out.halo
                halo_msg = f" for halo {out.halo}"
            else:
                halo = [0] * len(min_out_shape)
                halo_msg = ""

            if any([s - 2 * h < 1 for s, h in zip(min_out_shape, halo)]):
                raise ValueError(
                    f"Minimal shape {min_out_shape} of output {out.name} is too"
                    + f" small{halo_msg}."
                )

        return self

    @classmethod
    def _get_min_shape(
        cls,
        t: Union[InputTensorDescr, OutputTensorDescr],
        tensors_by_name: Dict[TensorName, Union[InputTensorDescr, OutputTensorDescr]],
    ) -> Sequence[int]:
        """output with subtracted halo has to result in meaningful output even for the minimal input
        see https://github.com/bioimage-io/spec-bioimage-io/issues/392
        """
        if isinstance(t.shape, collections.abc.Sequence):
            return t.shape
        elif isinstance(t.shape, ParameterizedInputShape):
            return t.shape.min
        elif isinstance(t.shape, ImplicitOutputShape):
            pass
        else:
            assert_never(t.shape)

        ref_shape = cls._get_min_shape(
            tensors_by_name[t.shape.reference_tensor], tensors_by_name
        )

        if None not in t.shape.scale:
            scale: Sequence[float, ...] = t.shape.scale  # type: ignore
        else:
            expanded_dims = [idx for idx, sc in enumerate(t.shape.scale) if sc is None]
            new_ref_shape: List[int] = []
            for idx in range(len(t.shape.scale)):
                ref_idx = idx - sum(int(exp < idx) for exp in expanded_dims)
                new_ref_shape.append(1 if idx in expanded_dims else ref_shape[ref_idx])

            ref_shape = new_ref_shape
            assert len(ref_shape) == len(t.shape.scale)
            scale = [0.0 if sc is None else sc for sc in t.shape.scale]

        offset = t.shape.offset
        assert len(offset) == len(scale)
        return [int(rs * s + 2 * off) for rs, s, off in zip(ref_shape, scale, offset)]

    @model_validator(mode="after")
    def validate_tensor_references_in_inputs(self) -> Self:
        for t in self.inputs:
            for proc in t.preprocessing:
                if "reference_tensor" not in proc.kwargs:
                    continue

                ref_tensor = proc.kwargs["reference_tensor"]
                if ref_tensor is not None and str(ref_tensor) not in {
                    str(t.name) for t in self.inputs
                }:
                    raise ValueError(f"'{ref_tensor}' not found in inputs")

                if ref_tensor == t.name:
                    raise ValueError(
                        f"invalid self reference for preprocessing of tensor {t.name}"
                    )

        return self

    @model_validator(mode="after")
    def validate_tensor_references_in_outputs(self) -> Self:
        for t in self.outputs:
            for proc in t.postprocessing:
                if "reference_tensor" not in proc.kwargs:
                    continue
                ref_tensor = proc.kwargs["reference_tensor"]
                if ref_tensor is not None and str(ref_tensor) not in {
                    str(t.name) for t in self.inputs
                }:
                    raise ValueError(f"{ref_tensor} not found in inputs")

        return self

    packaged_by: List[Author] = Field(
        default_factory=cast(Callable[[], List[Author]], list)
    )
    """The persons that have packaged and uploaded this model.
    Only required if those persons differ from the `authors`."""

    parent: Optional[LinkedModel] = None
    """The model from which this model is derived, e.g. by fine-tuning the weights."""

    @field_validator("parent", mode="before")
    @classmethod
    def ignore_url_parent(cls, parent: Any):
        if isinstance(parent, dict):
            return None

        else:
            return parent

    run_mode: Optional[RunMode] = None
    """Custom run mode for this model: for more complex prediction procedures like test time
    data augmentation that currently cannot be expressed in the specification.
    No standard run modes are defined yet."""

    sample_inputs: List[FileSource_] = Field(
        default_factory=cast(Callable[[], List[FileSource_]], list)
    )
    """URLs/relative paths to sample inputs to illustrate possible inputs for the model,
    for example stored as PNG or TIFF images.
    The sample files primarily serve to inform a human user about an example use case"""

    sample_outputs: List[FileSource_] = Field(
        default_factory=cast(Callable[[], List[FileSource_]], list)
    )
    """URLs/relative paths to sample outputs corresponding to the `sample_inputs`."""

    test_inputs: NotEmpty[
        List[Annotated[FileSource_, WithSuffix(".npy", case_sensitive=True)]]
    ]
    """Test input tensors compatible with the `inputs` description for a **single test case**.
    This means if your model has more than one input, you should provide one URL/relative path for each input.
    Each test input should be a file with an ndarray in
    [numpy.lib file format](https://numpy.org/doc/stable/reference/generated/numpy.lib.format.html#module-numpy.lib.format).
    The extension must be '.npy'."""

    test_outputs: NotEmpty[
        List[Annotated[FileSource_, WithSuffix(".npy", case_sensitive=True)]]
    ]
    """Analog to `test_inputs`."""

    timestamp: Datetime
    """Timestamp in [ISO 8601](#https://en.wikipedia.org/wiki/ISO_8601) format
    with a few restrictions listed [here](https://docs.python.org/3/library/datetime.html#datetime.datetime.fromisoformat)."""

    training_data: Union[LinkedDataset, DatasetDescr, None] = None
    """The dataset used to train this model"""

    weights: Annotated[WeightsDescr, WrapSerializer(package_weights)]
    """The weights for this model.
    Weights can be given for different formats, but should otherwise be equivalent.
    The available weight formats determine which consumers can use this model."""

    @model_validator(mode="before")
    @classmethod
    def _convert_from_older_format(
        cls, data: BioimageioYamlContent, /
    ) -> BioimageioYamlContent:
        convert_from_older_format(data)
        return data

    def get_input_test_arrays(self) -> List[NDArray[Any]]:
        data = [load_array(ipt) for ipt in self.test_inputs]
        assert all(isinstance(d, np.ndarray) for d in data)
        return data

    def get_output_test_arrays(self) -> List[NDArray[Any]]:
        data = [load_array(out) for out in self.test_outputs]
        assert all(isinstance(d, np.ndarray) for d in data)
        return data



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import (
    Any,
    Dict,
    Final,
    Generic,
    Type,
    Union,
    cast,
)

from typing_extensions import (
    TypeVar,
    TypeVarTuple,
    Unpack,
)

from .node import Node
from .utils import (
    assert_all_params_set_explicitly,
)
from .validated_string import ValidatedString

SRC = TypeVar("SRC", bound=Union[Node, ValidatedString])
TGT = TypeVar("TGT", bound=Node)


# converter without any additional args or kwargs:
# class Converter(Generic[SRC, TGT], ABC):
#     # src: ClassVar[Type[SRC]]
#     # tgt: ClassVar[Type[TGT]]
#     # note: the above is not yet possible, see https://github.com/python/typing/discussions/1424
#     # we therefore use an instance
#     def __init__(self, src: Type[SRC], tgt: Type[TGT], /):
#         super().__init__()
#         self.src: Final[Type[SRC]] = src
#         self.tgt: Final[Type[TGT]] = tgt

#     @abstractmethod
#     def _convert(self, src: SRC, tgt: "type[TGT | dict[str, Any]] ", /) -> "TGT | dict[str, Any]":
#         ...

#     def convert(self, source: SRC, /) -> TGT:
#         """convert `source` node

#         Args:
#             source: A bioimageio description node

#         Raises:
#             ValidationError: conversion failed
#         """
#         data = self.convert_as_dict(source)
#         return assert_all_params_set_explicitly(self.tgt)(**data)

#     def convert_as_dict(self, source: SRC) -> Dict[str, Any]:
#         return cast(Dict[str, Any], self._convert(source, dict))


# A TypeVar bound to a TypedDict seemed like a good way to add converter kwargs:
# ```
# class ConverterKwargs(TypedDict):
#     pass
# KW = TypeVar("KW", bound=ConverterKwargs, default=ConverterKwargs)
# ```
# sadly we cannot use a TypeVar bound to TypedDict and then unpack it in the Converter methods,
# see https://github.com/python/typing/issues/1399
# Therefore we use a TypeVarTuple and positional only args instead
# (We are avoiding ParamSpec for its ambiguity 'args vs kwargs')
CArgs = TypeVarTuple("CArgs")


class Converter(Generic[SRC, TGT, Unpack[CArgs]], ABC):
    # src: ClassVar[Type[SRC]]
    # tgt: ClassVar[Type[TGT]]
    # note: the above is not yet possible, see https://github.com/python/typing/discussions/1424
    # we therefore use an instance
    def __init__(self, src: Type[SRC], tgt: Type[TGT], /):
        super().__init__()
        self.src: Final[Type[SRC]] = src
        self.tgt: Final[Type[TGT]] = tgt

    @abstractmethod
    def _convert(
        self, src: SRC, tgt: "type[TGT | dict[str, Any]]", /, *args: Unpack[CArgs]
    ) -> "TGT | dict[str, Any]": ...

    # note: the following is not (yet) allowed, see https://github.com/python/typing/issues/1399
    #       we therefore use `kwargs` (and not `**kwargs`)
    # def convert(self, source: SRC, /, **kwargs: Unpack[KW]) -> TGT:
    def convert(self, source: SRC, /, *args: Unpack[CArgs]) -> TGT:
        """convert `source` node

        Args:
            source: A bioimageio description node

        Raises:
            ValidationError: conversion failed
        """
        data = self.convert_as_dict(source, *args)
        return assert_all_params_set_explicitly(self.tgt)(**data)

    def convert_as_dict(self, source: SRC, /, *args: Unpack[CArgs]) -> Dict[str, Any]:
        return cast(Dict[str, Any], self._convert(source, dict, *args))



"""use these type guards with caution!
They widen the type to T[Any], which is not always correct."""

import collections.abc
from typing import Any, Dict, List, Mapping, Sequence, Set, Tuple

import numpy as np
from numpy.typing import NDArray
from typing_extensions import TypeGuard


def is_dict(v: Any) -> TypeGuard[Dict[Any, Any]]:
    """to avoid Dict[Unknown, Unknown]"""
    return isinstance(v, dict)


def is_set(v: Any) -> TypeGuard[Set[Any]]:
    """to avoid Set[Unknown]"""
    return isinstance(v, set)


def is_kwargs(v: Any) -> TypeGuard[Dict[str, Any]]:
    return isinstance(v, dict) and all(
        isinstance(k, str)
        for k in v  # pyright: ignore[reportUnknownVariableType]
    )


def is_mapping(v: Any) -> TypeGuard[Mapping[Any, Any]]:
    """to avoid Mapping[Unknown, Unknown]"""
    return isinstance(v, collections.abc.Mapping)


def is_sequence(v: Any) -> TypeGuard[Sequence[Any]]:
    """to avoid Sequence[Unknown]"""
    return isinstance(v, collections.abc.Sequence)


def is_tuple(v: Any) -> TypeGuard[Tuple[Any, ...]]:
    """to avoid Tuple[Unknown, ...]"""
    return isinstance(v, tuple)


def is_list(v: Any) -> TypeGuard[List[Any]]:
    """to avoid List[Unknown]"""
    return isinstance(v, list)


def is_ndarray(v: Any) -> TypeGuard[NDArray[Any]]:
    return isinstance(v, np.ndarray)



import string
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Dict,
    List,
    Literal,
    Mapping,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
    cast,
)

import annotated_types
import pydantic
from annotated_types import Len, LowerCase, MaxLen
from pydantic import (
    EmailStr,
    Field,
    RootModel,
    ValidationInfo,
    field_validator,
    model_validator,
)
from typing_extensions import Annotated, Self, assert_never

from .._internal.common_nodes import Node, ResourceDescrBase
from .._internal.constants import TAG_CATEGORIES
from .._internal.field_warning import as_warning, issue_warning, warn
from .._internal.io import (
    BioimageioYamlContent,
    WithSuffix,
    YamlValue,
    wo_special_file_name,
)
from .._internal.io_packaging import FileSource_, include_in_package
from .._internal.type_guards import is_sequence
from .._internal.types import (
    DeprecatedLicenseId,
    FilePath,
    FileSource,
    LicenseId,
    NotEmpty,
)
from .._internal.types import Doi as Doi
from .._internal.types import OrcidId as OrcidId
from .._internal.types import RelativeFilePath as RelativeFilePath
from .._internal.url import HttpUrl as HttpUrl
from .._internal.validated_string import ValidatedString
from .._internal.validator_annotations import AfterValidator, RestrictCharacters
from .._internal.version_type import Version as Version
from ._v0_2_converter import convert_from_older_format as _convert_from_older_format


class ResourceId(ValidatedString):
    root_model: ClassVar[Type[RootModel[Any]]] = RootModel[
        Annotated[
            NotEmpty[str],
            AfterValidator(str.lower),  # convert upper case on the fly
            RestrictCharacters(string.ascii_lowercase + string.digits + "_-/."),
            annotated_types.Predicate(
                lambda s: not (s.startswith("/") or s.endswith("/"))
            ),
        ]
    ]


KNOWN_SPECIFIC_RESOURCE_TYPES = (
    "application",
    "collection",
    "dataset",
    "model",
    "notebook",
)

VALID_COVER_IMAGE_EXTENSIONS = (
    ".gif",
    ".jpeg",
    ".jpg",
    ".png",
    ".svg",
    ".tif",
    ".tiff",
)


FileSource_cover = Annotated[
    FileSource_,
    WithSuffix(VALID_COVER_IMAGE_EXTENSIONS, case_sensitive=False),
]


class AttachmentsDescr(Node):
    model_config = {**Node.model_config, "extra": "allow"}
    """update pydantic model config to allow additional unknown keys"""

    files: List[FileSource_] = Field(
        default_factory=cast(Callable[[], List[FileSource_]], list)
    )
    """File attachments"""


def _remove_slashes(s: str):
    return s.replace("/", "").replace("\\", "")


class Uploader(Node):
    email: EmailStr
    """Email"""
    name: Optional[Annotated[str, AfterValidator(_remove_slashes)]] = None
    """name"""


class _Person(Node):
    affiliation: Optional[str] = None
    """Affiliation"""

    email: Optional[EmailStr] = None
    """Email"""

    orcid: Annotated[Optional[OrcidId], Field(examples=["0000-0001-2345-6789"])] = None
    """An [ORCID iD](https://support.orcid.org/hc/en-us/sections/360001495313-What-is-ORCID
    ) in hyphenated groups of 4 digits, (and [valid](
    https://support.orcid.org/hc/en-us/articles/360006897674-Structure-of-the-ORCID-Identifier
    ) as per ISO 7064 11,2.)
    """


class Author(_Person):
    name: Annotated[str, AfterValidator(_remove_slashes)]
    github_user: Optional[str] = None  # TODO: validate github_user


class Maintainer(_Person):
    name: Optional[Annotated[str, AfterValidator(_remove_slashes)]] = None
    github_user: str


class BadgeDescr(Node):
    """A custom badge"""

    label: Annotated[str, Field(examples=["Open in Colab"])]
    """badge label to display on hover"""

    icon: Annotated[
        Optional[
            Union[
                Annotated[
                    Union[FilePath, RelativeFilePath],
                    AfterValidator(wo_special_file_name),
                    include_in_package,
                ],
                Union[HttpUrl, pydantic.HttpUrl],
            ]
        ],
        Field(examples=["https://colab.research.google.com/assets/colab-badge.svg"]),
    ] = None
    """badge icon (included in bioimage.io package if not a URL)"""

    url: Annotated[
        HttpUrl,
        Field(
            examples=[
                "https://colab.research.google.com/github/HenriquesLab/ZeroCostDL4Mic/blob/master/Colab_notebooks/U-net_2D_ZeroCostDL4Mic.ipynb"
            ]
        ),
    ]
    """target URL"""


class CiteEntry(Node):
    text: str
    """free text description"""

    doi: Optional[Doi] = None
    """A digital object identifier (DOI) is the prefered citation reference.
    See https://www.doi.org/ for details. (alternatively specify `url`)"""

    @field_validator("doi", mode="before")
    @classmethod
    def accept_prefixed_doi(cls, doi: Any) -> Any:
        if isinstance(doi, str):
            for doi_prefix in ("https://doi.org/", "http://dx.doi.org/"):
                if doi.startswith(doi_prefix):
                    doi = doi[len(doi_prefix) :]
                    break

        return doi

    url: Optional[str] = None
    """URL to cite (preferably specify a `doi` instead)"""

    @model_validator(mode="after")
    def _check_doi_or_url(self) -> Self:
        if not self.doi and not self.url:
            raise ValueError("Either 'doi' or 'url' is required")

        return self


class LinkedResource(Node):
    """Reference to a bioimage.io resource"""

    id: ResourceId
    """A valid resource `id` from the bioimage.io collection."""

    version_number: Optional[int] = None
    """version number (n-th published version, not the semantic version) of linked resource"""


class GenericModelDescrBase(ResourceDescrBase):
    """Base for all resource descriptions including of model descriptions"""

    name: Annotated[NotEmpty[str], warn(MaxLen(128), "Longer than 128 characters.")]
    """A human-friendly name of the resource description"""

    description: str

    covers: List[FileSource_cover] = Field(
        default_factory=cast(Callable[[], List[FileSource_cover]], list),
        examples=[["cover.png"]],
        description=(
            "Cover images. Please use an image smaller than 500KB and an aspect"
            " ratio width to height of 2:1.\nThe supported image formats are:"
            f" {VALID_COVER_IMAGE_EXTENSIONS}"
        ),
    )
    """Cover images. Please use an image smaller than 500KB and an aspect ratio width to height of 2:1."""

    id_emoji: Optional[
        Annotated[str, Len(min_length=1, max_length=1), Field(examples=["🦈", "🦥"])]
    ] = None
    """UTF-8 emoji for display alongside the `id`."""

    authors: List[Author] = Field(  # pyright: ignore[reportUnknownVariableType]
        default_factory=list
    )
    """The authors are the creators of the RDF and the primary points of contact."""

    @field_validator("authors", mode="before")
    @classmethod
    def accept_author_strings(cls, authors: Union[Any, Sequence[Any]]) -> Any:
        """we unofficially accept strings as author entries"""
        if is_sequence(authors):
            authors = [{"name": a} if isinstance(a, str) else a for a in authors]

        if not authors:
            issue_warning("missing", value=authors, field="authors")

        return authors

    attachments: Optional[AttachmentsDescr] = None
    """file and other attachments"""

    cite: List[CiteEntry] = Field(  # pyright: ignore[reportUnknownVariableType]
        default_factory=list
    )
    """citations"""

    @field_validator("cite", mode="after")
    @classmethod
    def _warn_empty_cite(cls, value: Any):
        if not value:
            issue_warning("missing", value=value, field="cite")

        return value

    config: Annotated[
        Dict[str, YamlValue],
        Field(
            examples=[
                dict(
                    bioimageio={
                        "my_custom_key": 3837283,
                        "another_key": {"nested": "value"},
                    },
                    imagej={"macro_dir": "path/to/macro/file"},
                )
            ],
        ),
    ] = Field(default_factory=dict)
    """A field for custom configuration that can contain any keys not present in the RDF spec.
    This means you should not store, for example, a github repo URL in `config` since we already have the
    `git_repo` field defined in the spec.
    Keys in `config` may be very specific to a tool or consumer software. To avoid conflicting definitions,
    it is recommended to wrap added configuration into a sub-field named with the specific domain or tool name,
    for example:
    ```yaml
    config:
        bioimageio:  # here is the domain name
            my_custom_key: 3837283
            another_key:
                nested: value
        imagej:       # config specific to ImageJ
            macro_dir: path/to/macro/file
    ```
    If possible, please use [`snake_case`](https://en.wikipedia.org/wiki/Snake_case) for keys in `config`.
    You may want to list linked files additionally under `attachments` to include them when packaging a resource
    (packaging a resource means downloading/copying important linked files and creating a ZIP archive that contains
    an altered rdf.yaml file with local references to the downloaded files)"""

    download_url: Optional[HttpUrl] = None
    """URL to download the resource from (deprecated)"""

    git_repo: Annotated[
        Optional[str],
        Field(
            examples=[
                "https://github.com/bioimage-io/spec-bioimage-io/tree/main/example_descriptions/models/unet2d_nuclei_broad"
            ],
        ),
    ] = None
    """A URL to the Git repository where the resource is being developed."""

    icon: Union[Annotated[str, Len(min_length=1, max_length=2)], FileSource, None] = (
        None
    )
    """An icon for illustration"""

    links: Annotated[
        List[str],
        Field(
            examples=[
                (
                    "ilastik/ilastik",
                    "deepimagej/deepimagej",
                    "zero/notebook_u-net_3d_zerocostdl4mic",
                )
            ],
        ),
    ] = Field(default_factory=list)
    """IDs of other bioimage.io resources"""

    uploader: Optional[Uploader] = None
    """The person who uploaded the model (e.g. to bioimage.io)"""

    # TODO: (py>3.8) remove pyright ignore
    maintainers: List[Maintainer] = Field(  # pyright: ignore[reportUnknownVariableType]
        default_factory=list
    )
    """Maintainers of this resource.
    If not specified `authors` are maintainers and at least some of them should specify their `github_user` name"""

    rdf_source: Optional[FileSource] = None
    """Resource description file (RDF) source; used to keep track of where an rdf.yaml was loaded from.
    Do not set this field in a YAML file."""

    tags: Annotated[
        List[str],
        Field(examples=[("unet2d", "pytorch", "nucleus", "segmentation", "dsb2018")]),
    ] = Field(default_factory=list)
    """Associated tags"""

    @as_warning
    @field_validator("tags")
    @classmethod
    def warn_about_tag_categories(
        cls, value: List[str], info: ValidationInfo
    ) -> List[str]:
        categories = TAG_CATEGORIES.get(info.data["type"], {})
        missing_categories: List[Mapping[str, Sequence[str]]] = []
        for cat, entries in categories.items():
            if not any(e in value for e in entries):
                missing_categories.append({cat: entries})

        if missing_categories:
            raise ValueError(
                "Missing tags from bioimage.io categories: {missing_categories}"
            )

        return value

    version: Optional[Version] = None
    """The version of the resource following SemVer 2.0."""

    version_number: Optional[int] = None
    """version number (n-th published version, not the semantic version)"""


class GenericDescrBase(GenericModelDescrBase):
    """Base for all resource descriptions except for the model descriptions"""

    implemented_format_version: ClassVar[Literal["0.2.4"]] = "0.2.4"
    if TYPE_CHECKING:
        format_version: Literal["0.2.4"] = "0.2.4"
    else:
        format_version: Literal["0.2.4"]
        """The format version of this resource specification
        (not the `version` of the resource description)
        When creating a new resource always use the latest micro/patch version described here.
        The `format_version` is important for any consumer software to understand how to parse the fields.
        """

    @model_validator(mode="before")
    @classmethod
    def _convert_from_older_format(
        cls, data: BioimageioYamlContent, /
    ) -> BioimageioYamlContent:
        _convert_from_older_format(data)
        return data

    badges: List[BadgeDescr] = Field(  # pyright: ignore[reportUnknownVariableType]
        default_factory=list
    )
    """badges associated with this resource"""

    documentation: Annotated[
        Optional[FileSource],
        Field(
            examples=[
                "https://raw.githubusercontent.com/bioimage-io/spec-bioimage-io/main/example_descriptions/models/unet2d_nuclei_broad/README.md",
                "README.md",
            ],
        ),
    ] = None
    """URL or relative path to a markdown file with additional documentation.
    The recommended documentation file name is `README.md`. An `.md` suffix is mandatory."""

    license: Annotated[
        Union[LicenseId, DeprecatedLicenseId, str, None],
        Field(union_mode="left_to_right", examples=["CC0-1.0", "MIT", "BSD-2-Clause"]),
    ] = None
    """A [SPDX license identifier](https://spdx.org/licenses/).
    We do not support custom license beyond the SPDX license list, if you need that please
    [open a GitHub issue](https://github.com/bioimage-io/spec-bioimage-io/issues/new/choose
    ) to discuss your intentions with the community."""

    @field_validator("license", mode="after")
    @classmethod
    def deprecated_spdx_license(
        cls, value: Optional[Union[LicenseId, DeprecatedLicenseId, str]]
    ):
        if isinstance(value, LicenseId):
            pass
        elif value is None:
            issue_warning("missing", value=value, field="license")
        elif isinstance(value, DeprecatedLicenseId):
            issue_warning(
                "'{value}' is a deprecated license identifier.",
                value=value,
                field="license",
            )
        elif isinstance(value, str):
            issue_warning(
                "'{value}' is an unknown license identifier.",
                value=value,
                field="license",
            )
        else:
            assert_never(value)

        return value


ResourceDescrType = TypeVar("ResourceDescrType", bound=GenericDescrBase)


class GenericDescr(GenericDescrBase, extra="ignore"):
    """Specification of the fields used in a generic bioimage.io-compliant resource description file (RDF).

    An RDF is a YAML file that describes a resource such as a model, a dataset, or a notebook.
    Note that those resources are described with a type-specific RDF.
    Use this generic resource description, if none of the known specific types matches your resource.
    """

    type: Annotated[str, LowerCase, Field(frozen=True)] = "generic"
    """The resource type assigns a broad category to the resource."""

    id: Optional[
        Annotated[ResourceId, Field(examples=["affable-shark", "ambitious-sloth"])]
    ] = None
    """bioimage.io-wide unique resource identifier
    assigned by bioimage.io; version **un**specific."""

    source: Optional[HttpUrl] = None
    """The primary source of the resource"""

    @field_validator("type", mode="after")
    @classmethod
    def check_specific_types(cls, value: str) -> str:
        if value in KNOWN_SPECIFIC_RESOURCE_TYPES:
            raise ValueError(
                f"Use the {value} description instead of this generic description for"
                + f" your '{value}' resource."
            )

        return value



import hashlib
import os
import zipfile
from contextlib import nullcontext
from functools import partial
from pathlib import Path
from typing import Any, ClassVar, Optional, Protocol, Type, Union, runtime_checkable
from zipfile import ZipFile

import pydantic
import zipp  # pyright: ignore[reportMissingTypeStubs]
from annotated_types import Predicate
from pydantic import RootModel, StringConstraints
from typing_extensions import Annotated

from .root_url import RootHttpUrl
from .validated_string import ValidatedString

FileName = str
FilePath = Annotated[pydantic.FilePath, pydantic.Field(title="FilePath")]
AbsoluteDirectory = Annotated[
    pydantic.DirectoryPath,
    Predicate(Path.is_absolute),
    pydantic.Field(title="AbsoluteDirectory"),
]
AbsoluteFilePath = Annotated[
    pydantic.FilePath,
    Predicate(Path.is_absolute),
    pydantic.Field(title="AbsoluteFilePath"),
]

BIOIMAGEIO_YAML = "rdf.yaml"
ALTERNATIVE_BIOIMAGEIO_YAML_NAMES = ("bioimageio.yaml", "model.yaml")
ALL_BIOIMAGEIO_YAML_NAMES = (BIOIMAGEIO_YAML,) + ALTERNATIVE_BIOIMAGEIO_YAML_NAMES

ZipPath = zipp.Path  # not zipfile.Path due to https://bugs.python.org/issue40564


class Sha256(ValidatedString):
    """A SHA-256 hash value"""

    root_model: ClassVar[Type[RootModel[Any]]] = RootModel[
        Annotated[
            str,
            StringConstraints(
                strip_whitespace=True, to_lower=True, min_length=64, max_length=64
            ),
        ]
    ]


class BytesReaderP(Protocol):
    def read(self, size: int = -1, /) -> bytes: ...

    @property
    def closed(self) -> bool: ...

    def readable(self) -> bool: ...

    def seek(self, offset: int, whence: int = os.SEEK_SET, /) -> int: ...

    def seekable(self) -> bool: ...

    def tell(self) -> int: ...


@runtime_checkable
class BytesReaderIntoP(BytesReaderP, Protocol):
    def readinto(self, b: Union[bytearray, memoryview]) -> int: ...


Suffix = str


class BytesReader(BytesReaderP):
    def __init__(
        self,
        /,
        reader: Union[BytesReaderP, BytesReaderIntoP],
        *,
        sha256: Optional[Sha256],
        suffix: Suffix,
        original_file_name: FileName,
        original_root: Union[RootHttpUrl, AbsoluteDirectory, ZipFile],
        is_zipfile: Optional[bool],
    ) -> None:
        self._reader = reader
        self._sha256 = sha256
        self._suffix = suffix
        self._original_file_name = original_file_name
        self._original_root = original_root
        self._is_zipfile = is_zipfile
        super().__init__()

    @property
    def is_zipfile(self) -> bool:
        if self._is_zipfile is None:
            pos = self.tell()
            self._is_zipfile = zipfile.is_zipfile(self)
            _ = self.seek(pos)

        return self._is_zipfile

    @property
    def sha256(self) -> Sha256:
        if self._sha256 is None:
            pos = self._reader.tell()
            _ = self._reader.seek(0)
            self._sha256 = get_sha256(self._reader)
            _ = self._reader.seek(pos)

        return self._sha256

    @property
    def suffix(self) -> Suffix:
        return self._suffix

    @property
    def original_file_name(self) -> FileName:
        return self._original_file_name

    @property
    def original_root(self) -> Union[RootHttpUrl, AbsoluteDirectory, ZipFile]:
        return self._original_root

    def read(self, size: int = -1, /) -> bytes:
        return self._reader.read(size)

    def read_text(self, encoding: str = "utf-8") -> str:
        return self._reader.read().decode(encoding)

    def readable(self) -> bool:
        return True

    def seek(self, offset: int, whence: int = os.SEEK_SET, /) -> int:
        return self._reader.seek(offset, whence)

    def seekable(self) -> bool:
        return True

    def tell(self) -> int:
        return self._reader.tell()

    @property
    def closed(self) -> bool:
        return self._reader.closed


def get_sha256(source: Union[BytesReaderP, BytesReaderIntoP, Path]) -> Sha256:
    chunksize = 128 * 1024
    h = hashlib.sha256()

    if isinstance(source, BytesReaderIntoP):
        b = bytearray(chunksize)
        mv = memoryview(b)
        for n in iter(lambda: source.readinto(mv), 0):
            h.update(mv[:n])
    else:
        if isinstance(source, Path):
            read_ctxt = source.open(mode="rb")
        else:
            read_ctxt = nullcontext(source)

        with read_ctxt as r:
            for chunk in iter(partial(r.read, chunksize), b""):
                h.update(chunk)

    sha = h.hexdigest()
    return Sha256(sha)



from __future__ import annotations

import warnings
from pathlib import Path
from typing import (
    Optional,
    Union,
)

from pydantic import (
    AnyUrl,
    SerializationInfo,
    SerializerFunctionWrapHandler,
    WrapSerializer,
)
from typing_extensions import (
    Annotated,
    assert_never,
)

from .io import (
    FileDescr,
    FileSource,
    RelativeFilePath,
    extract_file_name,
    wo_special_file_name,
)
from .io_basics import (
    ALL_BIOIMAGEIO_YAML_NAMES,
    FileName,
    Sha256,
)
from .packaging_context import packaging_context_var
from .url import HttpUrl
from .utils import PrettyPlainSerializer
from .validator_annotations import AfterValidator


def _package_serializer(
    source: FileSource, info: SerializationInfo
) -> Union[str, Path, FileName]:
    return _package(source, info, None)


def package_file_descr_serializer(
    value: FileDescr, handler: SerializerFunctionWrapHandler, info: SerializationInfo
):
    ret = handler(
        value,
        info,  # pyright: ignore[reportArgumentType]  # taken from pydantic docs
    )
    ret["source"] = _package(value.source, info, sha256=value.sha256)
    return ret


def _package(
    source: FileSource, info: SerializationInfo, sha256: Optional[Sha256]
) -> Union[str, Path, FileName]:
    if (packaging_context := packaging_context_var.get()) is None:
        # convert to standard python obj
        # note: pydantic keeps returning Rootmodels (here `HttpUrl`) as-is, but if
        #   this function returns one RootModel, paths are "further serialized" by
        #   returning the 'root' attribute, which is incorrect.
        #   see https://github.com/pydantic/pydantic/issues/8963
        #   TODO: follow up on https://github.com/pydantic/pydantic/issues/8963
        if isinstance(source, Path):
            unpackaged = source
        elif isinstance(source, HttpUrl):
            unpackaged = source
        elif isinstance(source, RelativeFilePath):
            unpackaged = Path(source.path)
        elif isinstance(source, AnyUrl):
            unpackaged = str(source)
        else:
            assert_never(source)

        if info.mode_is_json():
            # convert to json value  # TODO: remove and let pydantic do this?
            if isinstance(unpackaged, (Path, HttpUrl)):
                unpackaged = str(unpackaged)
            elif isinstance(unpackaged, str):
                pass
            else:
                assert_never(unpackaged)
        else:
            warnings.warn(
                "dumping with mode='python' is currently not fully supported for "
                + "fields that are included when packaging; returned objects are "
                + "standard python objects"
            )

        return unpackaged  # return unpackaged file source

    fname = extract_file_name(source)
    if not fname:
        raise ValueError(f"Got empty file name for source {source}")

    if fname == packaging_context.bioimageio_yaml_file_name:
        raise ValueError(
            f"Reserved file name '{packaging_context.bioimageio_yaml_file_name}' "
            + "not allowed for a file to be packaged"
        )

    fsrcs = packaging_context.file_sources
    assert not any(fname.endswith(special) for special in ALL_BIOIMAGEIO_YAML_NAMES), (
        fname
    )
    if fname in fsrcs and fsrcs[fname].source != source:
        for i in range(2, 20):
            fn, *ext = fname.split(".")
            alternative_file_name = ".".join([f"{fn}_{i}", *ext])
            if (
                alternative_file_name not in fsrcs
                or fsrcs[alternative_file_name].source == source
            ):
                fname = alternative_file_name
                break
        else:
            raise ValueError(f"Too many file name clashes for {fname}")

    fsrcs[fname] = FileDescr(source=source, sha256=sha256)
    return fname


include_in_package = PrettyPlainSerializer(_package_serializer, when_used="unless-none")
"""DEPRECATED serializer for `source` fields without corresponding `sha256` field."""

include_when_packaging = WrapSerializer(
    package_file_descr_serializer, when_used="unless-none"
)
"""Pydantic serializer that marks the annotated `FileDescr` to be included when packaging
(saving a bioimageio zip package)."""

FileSource_ = Annotated[
    FileSource,
    AfterValidator(wo_special_file_name),
    include_in_package,
]
"""A file source that is included when packaging the resource."""

FileDescr_ = Annotated[
    FileDescr, AfterValidator(wo_special_file_name), include_when_packaging
]
"""A `FileDescr` whose **source** is included when packaging the resource."""



from typing import Any, Collection, Dict, Iterable, Mapping, Tuple

import httpx
import pytest

from bioimageio.spec.common import HttpUrl, Sha256
from tests.utils import ParameterSet, check_bioimageio_yaml, expensive_test

BASE_URL = "https://uk1s3.embassy.ebi.ac.uk/public-datasets/bioimage.io/"

KNOWN_INVALID: Mapping[str, str] = {
    "stupendous-sheep/1.1": "requires relativ import of attachment",
    "whimsical-helmet/2.1.2": "invalid id",
    "modest-spider/0.1.1": "non-batch id 'b'",
}
EXCLUDE_FIELDS_FROM_ROUNDTRIP_DEFAULT: Collection[str] = {
    "version_number",  # deprecated field that gets dropped in favor of `version``
    "version",  # may be set from deprecated `version_number`
}
EXCLUDE_FIELDS_FROM_ROUNDTRIP: Mapping[str, Collection[str]] = {
    "affable-shark/1.1": {"inputs"},  # preprocessing ensure_dtype added
    "affable-shark/1.2": {"inputs"},  # preprocessing ensure_dtype added
    "charismatic-whale/1.0.1": {"inputs", "outputs"},  # int -> float
    "dynamic-t-rex/1": {"inputs"},  # int -> float
    "impartial-shrimp/1.1": {"inputs", "cite"},  # preprocessing ensure_dtype added
    "philosophical-panda/0.0.11": {"outputs"},  # int -> float
    "philosophical-panda/0.1.0": {"outputs"},  # int -> float
    "polite-pig/1.1": {"inputs", "outputs"},
    "charismatic-whale/1.0.2": {"inputs", "outputs"},  # int -> float
    "dynamic-t-rex/1.1": {"inputs"},  # int -> float
}


def _get_rdf_sources(only_latest: bool = True):
    entries: Any = httpx.get(BASE_URL + "all_versions.json").json()["entries"]
    ret: Dict[str, Tuple[HttpUrl, Sha256]] = {}
    for entry in entries:
        for version in entry["versions"]:
            ret[f"{entry['concept']}/{version['v']}"] = (
                HttpUrl(version["source"]),
                Sha256(version["sha256"]),
            )
            if only_latest:
                break

    return ret


ALL_RDF_SOURCES: Mapping[str, Tuple[HttpUrl, Sha256]] = _get_rdf_sources()


def yield_bioimageio_yaml_urls() -> Iterable[ParameterSet]:
    for descr_url, sha in ALL_RDF_SOURCES.values():
        key = (
            descr_url.replace(BASE_URL, "")
            .replace("/files/rdf.yaml", "")
            .replace("/files/bioimageio.yaml", "")
        )
        yield pytest.param(descr_url, sha, key, id=key)


@expensive_test
@pytest.mark.parametrize(
    "descr_url,sha,key",
    sorted(
        yield_bioimageio_yaml_urls(),
        key=lambda p: p.id,  # pyright: ignore[reportAttributeAccessIssue,reportUnknownLambdaType]
    ),
)
def test_rdf(
    descr_url: HttpUrl,
    sha: Sha256,
    key: str,
    bioimageio_json_schema: Mapping[Any, Any],
):
    if key in KNOWN_INVALID:
        pytest.skip(KNOWN_INVALID[key])

    check_bioimageio_yaml(
        descr_url,
        sha=sha,
        as_latest=False,
        exclude_fields_from_roundtrip=EXCLUDE_FIELDS_FROM_ROUNDTRIP.get(
            key, EXCLUDE_FIELDS_FROM_ROUNDTRIP_DEFAULT
        ),
        bioimageio_json_schema=bioimageio_json_schema,
        perform_io_checks=False,
    )


@expensive_test
@pytest.mark.parametrize(
    "rdf_id",
    [
        "ambitious-sloth/1.2",
        "breezy-handbag/1",
        "faithful-chicken/1.2",
        "ilastik/ilastik/1",
        "uplifting-ice-cream/1",
    ],
)
def test_exemplary_rdf(rdf_id: str, bioimageio_json_schema: Mapping[Any, Any]):
    """test a list of models we expect to be compatible with the latest spec version"""
    source, sha = ALL_RDF_SOURCES[rdf_id]
    check_bioimageio_yaml(
        source,
        sha=sha,
        as_latest=True,
        exclude_fields_from_roundtrip=EXCLUDE_FIELDS_FROM_ROUNDTRIP.get(
            rdf_id, EXCLUDE_FIELDS_FROM_ROUNDTRIP_DEFAULT
        ),
        bioimageio_json_schema=bioimageio_json_schema,
        perform_io_checks=True,
    )



from __future__ import annotations

from typing import Any, ClassVar, Iterable, Optional, Type
from urllib.parse import urlsplit, urlunsplit

import pydantic
from pydantic import RootModel

from .validated_string import ValidatedString


class RootHttpUrl(ValidatedString):
    """An untested HTTP URL, possibly a 'URL folder' or an invalid HTTP URL"""

    root_model: ClassVar[Type[RootModel[Any]]] = RootModel[pydantic.HttpUrl]
    _validated: pydantic.HttpUrl

    def absolute(self):
        """analog to `absolute` method of pathlib."""
        return self

    @property
    def scheme(self) -> str:
        return self._validated.scheme

    @property
    def host(self) -> Optional[str]:
        return self._validated.host

    @property
    def path(self) -> Optional[str]:
        return self._validated.path

    @property
    def parent(self) -> RootHttpUrl:
        parsed = urlsplit(str(self))
        path = list(parsed.path.split("/"))
        if (
            parsed.netloc == "zenodo.org"
            and parsed.path.startswith("/api/records/")
            and parsed.path.endswith("/content")
        ):
            path[-2:-1] = []
        else:
            path = path[:-1]

        return RootHttpUrl(
            urlunsplit(
                (
                    parsed.scheme,
                    parsed.netloc,
                    "/".join(path),
                    parsed.query,
                    parsed.fragment,
                )
            )
        )

    @property
    def parents(self) -> Iterable[RootHttpUrl]:
        """iterate over all URL parents (max 100)"""
        current = self
        for _ in range(100):
            current = current.parent
            yield current



# This file was generated by scripts/generate_spdx_licenses.py
from typing import Literal

LicenseId = Literal[
    "0BSD",
    "3D-Slicer-1.0",
    "AAL",
    "Abstyles",
    "AdaCore-doc",
    "Adobe-2006",
    "Adobe-Display-PostScript",
    "Adobe-Glyph",
    "Adobe-Utopia",
    "ADSL",
    "AFL-1.1",
    "AFL-1.2",
    "AFL-2.0",
    "AFL-2.1",
    "AFL-3.0",
    "Afmparse",
    "AGPL-1.0-only",
    "AGPL-1.0-or-later",
    "AGPL-3.0-only",
    "AGPL-3.0-or-later",
    "Aladdin",
    "AMD-newlib",
    "AMDPLPA",
    "AML",
    "AML-glslang",
    "AMPAS",
    "ANTLR-PD",
    "ANTLR-PD-fallback",
    "any-OSI",
    "any-OSI-perl-modules",
    "Apache-1.0",
    "Apache-1.1",
    "Apache-2.0",
    "APAFML",
    "APL-1.0",
    "App-s2p",
    "APSL-1.0",
    "APSL-1.1",
    "APSL-1.2",
    "APSL-2.0",
    "Arphic-1999",
    "Artistic-1.0",
    "Artistic-1.0-cl8",
    "Artistic-1.0-Perl",
    "Artistic-2.0",
    "Artistic-dist",
    "Aspell-RU",
    "ASWF-Digital-Assets-1.0",
    "ASWF-Digital-Assets-1.1",
    "Baekmuk",
    "Bahyph",
    "Barr",
    "bcrypt-Solar-Designer",
    "Beerware",
    "Bitstream-Charter",
    "Bitstream-Vera",
    "BitTorrent-1.0",
    "BitTorrent-1.1",
    "blessing",
    "BlueOak-1.0.0",
    "Boehm-GC",
    "Boehm-GC-without-fee",
    "Borceux",
    "Brian-Gladman-2-Clause",
    "Brian-Gladman-3-Clause",
    "BSD-1-Clause",
    "BSD-2-Clause",
    "BSD-2-Clause-Darwin",
    "BSD-2-Clause-first-lines",
    "BSD-2-Clause-Patent",
    "BSD-2-Clause-pkgconf-disclaimer",
    "BSD-2-Clause-Views",
    "BSD-3-Clause",
    "BSD-3-Clause-acpica",
    "BSD-3-Clause-Attribution",
    "BSD-3-Clause-Clear",
    "BSD-3-Clause-flex",
    "BSD-3-Clause-HP",
    "BSD-3-Clause-LBNL",
    "BSD-3-Clause-Modification",
    "BSD-3-Clause-No-Military-License",
    "BSD-3-Clause-No-Nuclear-License",
    "BSD-3-Clause-No-Nuclear-License-2014",
    "BSD-3-Clause-No-Nuclear-Warranty",
    "BSD-3-Clause-Open-MPI",
    "BSD-3-Clause-Sun",
    "BSD-4-Clause",
    "BSD-4-Clause-Shortened",
    "BSD-4-Clause-UC",
    "BSD-4.3RENO",
    "BSD-4.3TAHOE",
    "BSD-Advertising-Acknowledgement",
    "BSD-Attribution-HPND-disclaimer",
    "BSD-Inferno-Nettverk",
    "BSD-Protection",
    "BSD-Source-beginning-file",
    "BSD-Source-Code",
    "BSD-Systemics",
    "BSD-Systemics-W3Works",
    "BSL-1.0",
    "BUSL-1.1",
    "bzip2-1.0.6",
    "C-UDA-1.0",
    "CAL-1.0",
    "CAL-1.0-Combined-Work-Exception",
    "Caldera",
    "Caldera-no-preamble",
    "Catharon",
    "CATOSL-1.1",
    "CC-BY-1.0",
    "CC-BY-2.0",
    "CC-BY-2.5",
    "CC-BY-2.5-AU",
    "CC-BY-3.0",
    "CC-BY-3.0-AT",
    "CC-BY-3.0-AU",
    "CC-BY-3.0-DE",
    "CC-BY-3.0-IGO",
    "CC-BY-3.0-NL",
    "CC-BY-3.0-US",
    "CC-BY-4.0",
    "CC-BY-NC-1.0",
    "CC-BY-NC-2.0",
    "CC-BY-NC-2.5",
    "CC-BY-NC-3.0",
    "CC-BY-NC-3.0-DE",
    "CC-BY-NC-4.0",
    "CC-BY-NC-ND-1.0",
    "CC-BY-NC-ND-2.0",
    "CC-BY-NC-ND-2.5",
    "CC-BY-NC-ND-3.0",
    "CC-BY-NC-ND-3.0-DE",
    "CC-BY-NC-ND-3.0-IGO",
    "CC-BY-NC-ND-4.0",
    "CC-BY-NC-SA-1.0",
    "CC-BY-NC-SA-2.0",
    "CC-BY-NC-SA-2.0-DE",
    "CC-BY-NC-SA-2.0-FR",
    "CC-BY-NC-SA-2.0-UK",
    "CC-BY-NC-SA-2.5",
    "CC-BY-NC-SA-3.0",
    "CC-BY-NC-SA-3.0-DE",
    "CC-BY-NC-SA-3.0-IGO",
    "CC-BY-NC-SA-4.0",
    "CC-BY-ND-1.0",
    "CC-BY-ND-2.0",
    "CC-BY-ND-2.5",
    "CC-BY-ND-3.0",
    "CC-BY-ND-3.0-DE",
    "CC-BY-ND-4.0",
    "CC-BY-SA-1.0",
    "CC-BY-SA-2.0",
    "CC-BY-SA-2.0-UK",
    "CC-BY-SA-2.1-JP",
    "CC-BY-SA-2.5",
    "CC-BY-SA-3.0",
    "CC-BY-SA-3.0-AT",
    "CC-BY-SA-3.0-DE",
    "CC-BY-SA-3.0-IGO",
    "CC-BY-SA-4.0",
    "CC-PDDC",
    "CC-PDM-1.0",
    "CC-SA-1.0",
    "CC0-1.0",
    "CDDL-1.0",
    "CDDL-1.1",
    "CDL-1.0",
    "CDLA-Permissive-1.0",
    "CDLA-Permissive-2.0",
    "CDLA-Sharing-1.0",
    "CECILL-1.0",
    "CECILL-1.1",
    "CECILL-2.0",
    "CECILL-2.1",
    "CECILL-B",
    "CECILL-C",
    "CERN-OHL-1.1",
    "CERN-OHL-1.2",
    "CERN-OHL-P-2.0",
    "CERN-OHL-S-2.0",
    "CERN-OHL-W-2.0",
    "CFITSIO",
    "check-cvs",
    "checkmk",
    "ClArtistic",
    "Clips",
    "CMU-Mach",
    "CMU-Mach-nodoc",
    "CNRI-Jython",
    "CNRI-Python",
    "CNRI-Python-GPL-Compatible",
    "COIL-1.0",
    "Community-Spec-1.0",
    "Condor-1.1",
    "copyleft-next-0.3.0",
    "copyleft-next-0.3.1",
    "Cornell-Lossless-JPEG",
    "CPAL-1.0",
    "CPL-1.0",
    "CPOL-1.02",
    "Cronyx",
    "Crossword",
    "CryptoSwift",
    "CrystalStacker",
    "CUA-OPL-1.0",
    "Cube",
    "curl",
    "cve-tou",
    "D-FSL-1.0",
    "DEC-3-Clause",
    "diffmark",
    "DL-DE-BY-2.0",
    "DL-DE-ZERO-2.0",
    "DOC",
    "DocBook-DTD",
    "DocBook-Schema",
    "DocBook-Stylesheet",
    "DocBook-XML",
    "Dotseqn",
    "DRL-1.0",
    "DRL-1.1",
    "DSDP",
    "dtoa",
    "dvipdfm",
    "ECL-1.0",
    "ECL-2.0",
    "EFL-1.0",
    "EFL-2.0",
    "eGenix",
    "Elastic-2.0",
    "Entessa",
    "EPICS",
    "EPL-1.0",
    "EPL-2.0",
    "ErlPL-1.1",
    "etalab-2.0",
    "EUDatagrid",
    "EUPL-1.0",
    "EUPL-1.1",
    "EUPL-1.2",
    "Eurosym",
    "Fair",
    "FBM",
    "FDK-AAC",
    "Ferguson-Twofish",
    "Frameworx-1.0",
    "FreeBSD-DOC",
    "FreeImage",
    "FSFAP",
    "FSFAP-no-warranty-disclaimer",
    "FSFUL",
    "FSFULLR",
    "FSFULLRSD",
    "FSFULLRWD",
    "FSL-1.1-ALv2",
    "FSL-1.1-MIT",
    "FTL",
    "Furuseth",
    "fwlw",
    "Game-Programming-Gems",
    "GCR-docs",
    "GD",
    "generic-xts",
    "GFDL-1.1-invariants-only",
    "GFDL-1.1-invariants-or-later",
    "GFDL-1.1-no-invariants-only",
    "GFDL-1.1-no-invariants-or-later",
    "GFDL-1.1-only",
    "GFDL-1.1-or-later",
    "GFDL-1.2-invariants-only",
    "GFDL-1.2-invariants-or-later",
    "GFDL-1.2-no-invariants-only",
    "GFDL-1.2-no-invariants-or-later",
    "GFDL-1.2-only",
    "GFDL-1.2-or-later",
    "GFDL-1.3-invariants-only",
    "GFDL-1.3-invariants-or-later",
    "GFDL-1.3-no-invariants-only",
    "GFDL-1.3-no-invariants-or-later",
    "GFDL-1.3-only",
    "GFDL-1.3-or-later",
    "Giftware",
    "GL2PS",
    "Glide",
    "Glulxe",
    "GLWTPL",
    "gnuplot",
    "GPL-1.0-only",
    "GPL-1.0-or-later",
    "GPL-2.0-only",
    "GPL-2.0-or-later",
    "GPL-3.0-only",
    "GPL-3.0-or-later",
    "Graphics-Gems",
    "gSOAP-1.3b",
    "gtkbook",
    "Gutmann",
    "HaskellReport",
    "HDF5",
    "hdparm",
    "HIDAPI",
    "Hippocratic-2.1",
    "HP-1986",
    "HP-1989",
    "HPND",
    "HPND-DEC",
    "HPND-doc",
    "HPND-doc-sell",
    "HPND-export-US",
    "HPND-export-US-acknowledgement",
    "HPND-export-US-modify",
    "HPND-export2-US",
    "HPND-Fenneberg-Livingston",
    "HPND-INRIA-IMAG",
    "HPND-Intel",
    "HPND-Kevlin-Henney",
    "HPND-Markus-Kuhn",
    "HPND-merchantability-variant",
    "HPND-MIT-disclaimer",
    "HPND-Netrek",
    "HPND-Pbmplus",
    "HPND-sell-MIT-disclaimer-xserver",
    "HPND-sell-regexpr",
    "HPND-sell-variant",
    "HPND-sell-variant-MIT-disclaimer",
    "HPND-sell-variant-MIT-disclaimer-rev",
    "HPND-UC",
    "HPND-UC-export-US",
    "HTMLTIDY",
    "IBM-pibs",
    "ICU",
    "IEC-Code-Components-EULA",
    "IJG",
    "IJG-short",
    "ImageMagick",
    "iMatix",
    "Imlib2",
    "Info-ZIP",
    "Inner-Net-2.0",
    "InnoSetup",
    "Intel",
    "Intel-ACPI",
    "Interbase-1.0",
    "IPA",
    "IPL-1.0",
    "ISC",
    "ISC-Veillard",
    "Jam",
    "JasPer-2.0",
    "jove",
    "JPL-image",
    "JPNIC",
    "JSON",
    "Kastrup",
    "Kazlib",
    "Knuth-CTAN",
    "LAL-1.2",
    "LAL-1.3",
    "Latex2e",
    "Latex2e-translated-notice",
    "Leptonica",
    "LGPL-2.0-only",
    "LGPL-2.0-or-later",
    "LGPL-2.1-only",
    "LGPL-2.1-or-later",
    "LGPL-3.0-only",
    "LGPL-3.0-or-later",
    "LGPLLR",
    "Libpng",
    "libpng-1.6.35",
    "libpng-2.0",
    "libselinux-1.0",
    "libtiff",
    "libutil-David-Nugent",
    "LiLiQ-P-1.1",
    "LiLiQ-R-1.1",
    "LiLiQ-Rplus-1.1",
    "Linux-man-pages-1-para",
    "Linux-man-pages-copyleft",
    "Linux-man-pages-copyleft-2-para",
    "Linux-man-pages-copyleft-var",
    "Linux-OpenIB",
    "LOOP",
    "LPD-document",
    "LPL-1.0",
    "LPL-1.02",
    "LPPL-1.0",
    "LPPL-1.1",
    "LPPL-1.2",
    "LPPL-1.3a",
    "LPPL-1.3c",
    "lsof",
    "Lucida-Bitmap-Fonts",
    "LZMA-SDK-9.11-to-9.20",
    "LZMA-SDK-9.22",
    "Mackerras-3-Clause",
    "Mackerras-3-Clause-acknowledgment",
    "magaz",
    "mailprio",
    "MakeIndex",
    "man2html",
    "Martin-Birgmeier",
    "McPhee-slideshow",
    "metamail",
    "Minpack",
    "MIPS",
    "MirOS",
    "MIT",
    "MIT-0",
    "MIT-advertising",
    "MIT-Click",
    "MIT-CMU",
    "MIT-enna",
    "MIT-feh",
    "MIT-Festival",
    "MIT-Khronos-old",
    "MIT-Modern-Variant",
    "MIT-open-group",
    "MIT-testregex",
    "MIT-Wu",
    "MITNFA",
    "MMIXware",
    "Motosoto",
    "MPEG-SSG",
    "mpi-permissive",
    "mpich2",
    "MPL-1.0",
    "MPL-1.1",
    "MPL-2.0",
    "MPL-2.0-no-copyleft-exception",
    "mplus",
    "MS-LPL",
    "MS-PL",
    "MS-RL",
    "MTLL",
    "MulanPSL-1.0",
    "MulanPSL-2.0",
    "Multics",
    "Mup",
    "NAIST-2003",
    "NASA-1.3",
    "Naumen",
    "NBPL-1.0",
    "NCBI-PD",
    "NCGL-UK-2.0",
    "NCL",
    "NCSA",
    "NetCDF",
    "Newsletr",
    "NGPL",
    "ngrep",
    "NICTA-1.0",
    "NIST-PD",
    "NIST-PD-fallback",
    "NIST-Software",
    "NLOD-1.0",
    "NLOD-2.0",
    "NLPL",
    "Nokia",
    "NOSL",
    "Noweb",
    "NPL-1.0",
    "NPL-1.1",
    "NPOSL-3.0",
    "NRL",
    "NTIA-PD",
    "NTP",
    "NTP-0",
    "O-UDA-1.0",
    "OAR",
    "OCCT-PL",
    "OCLC-2.0",
    "ODbL-1.0",
    "ODC-By-1.0",
    "OFFIS",
    "OFL-1.0",
    "OFL-1.0-no-RFN",
    "OFL-1.0-RFN",
    "OFL-1.1",
    "OFL-1.1-no-RFN",
    "OFL-1.1-RFN",
    "OGC-1.0",
    "OGDL-Taiwan-1.0",
    "OGL-Canada-2.0",
    "OGL-UK-1.0",
    "OGL-UK-2.0",
    "OGL-UK-3.0",
    "OGTSL",
    "OLDAP-1.1",
    "OLDAP-1.2",
    "OLDAP-1.3",
    "OLDAP-1.4",
    "OLDAP-2.0",
    "OLDAP-2.0.1",
    "OLDAP-2.1",
    "OLDAP-2.2",
    "OLDAP-2.2.1",
    "OLDAP-2.2.2",
    "OLDAP-2.3",
    "OLDAP-2.4",
    "OLDAP-2.5",
    "OLDAP-2.6",
    "OLDAP-2.7",
    "OLDAP-2.8",
    "OLFL-1.3",
    "OML",
    "OpenPBS-2.3",
    "OpenSSL",
    "OpenSSL-standalone",
    "OpenVision",
    "OPL-1.0",
    "OPL-UK-3.0",
    "OPUBL-1.0",
    "OSET-PL-2.1",
    "OSL-1.0",
    "OSL-1.1",
    "OSL-2.0",
    "OSL-2.1",
    "OSL-3.0",
    "PADL",
    "Parity-6.0.0",
    "Parity-7.0.0",
    "PDDL-1.0",
    "PHP-3.0",
    "PHP-3.01",
    "Pixar",
    "pkgconf",
    "Plexus",
    "pnmstitch",
    "PolyForm-Noncommercial-1.0.0",
    "PolyForm-Small-Business-1.0.0",
    "PostgreSQL",
    "PPL",
    "PSF-2.0",
    "psfrag",
    "psutils",
    "Python-2.0",
    "Python-2.0.1",
    "python-ldap",
    "Qhull",
    "QPL-1.0",
    "QPL-1.0-INRIA-2004",
    "radvd",
    "Rdisc",
    "RHeCos-1.1",
    "RPL-1.1",
    "RPL-1.5",
    "RPSL-1.0",
    "RSA-MD",
    "RSCPL",
    "Ruby",
    "Ruby-pty",
    "SAX-PD",
    "SAX-PD-2.0",
    "Saxpath",
    "SCEA",
    "SchemeReport",
    "Sendmail",
    "Sendmail-8.23",
    "Sendmail-Open-Source-1.1",
    "SGI-B-1.0",
    "SGI-B-1.1",
    "SGI-B-2.0",
    "SGI-OpenGL",
    "SGP4",
    "SHL-0.5",
    "SHL-0.51",
    "SimPL-2.0",
    "SISSL",
    "SISSL-1.2",
    "SL",
    "Sleepycat",
    "SMAIL-GPL",
    "SMLNJ",
    "SMPPL",
    "SNIA",
    "snprintf",
    "SOFA",
    "softSurfer",
    "Soundex",
    "Spencer-86",
    "Spencer-94",
    "Spencer-99",
    "SPL-1.0",
    "ssh-keyscan",
    "SSH-OpenSSH",
    "SSH-short",
    "SSLeay-standalone",
    "SSPL-1.0",
    "SugarCRM-1.1.3",
    "SUL-1.0",
    "Sun-PPP",
    "Sun-PPP-2000",
    "SunPro",
    "SWL",
    "swrule",
    "Symlinks",
    "TAPR-OHL-1.0",
    "TCL",
    "TCP-wrappers",
    "TermReadKey",
    "TGPPL-1.0",
    "ThirdEye",
    "threeparttable",
    "TMate",
    "TORQUE-1.1",
    "TOSL",
    "TPDL",
    "TPL-1.0",
    "TrustedQSL",
    "TTWL",
    "TTYP0",
    "TU-Berlin-1.0",
    "TU-Berlin-2.0",
    "Ubuntu-font-1.0",
    "UCAR",
    "UCL-1.0",
    "ulem",
    "UMich-Merit",
    "Unicode-3.0",
    "Unicode-DFS-2015",
    "Unicode-DFS-2016",
    "Unicode-TOU",
    "UnixCrypt",
    "Unlicense",
    "Unlicense-libtelnet",
    "Unlicense-libwhirlpool",
    "UPL-1.0",
    "URT-RLE",
    "Vim",
    "VOSTROM",
    "VSL-1.0",
    "W3C",
    "W3C-19980720",
    "W3C-20150513",
    "w3m",
    "Watcom-1.0",
    "Widget-Workshop",
    "Wsuipa",
    "WTFPL",
    "wwl",
    "X11",
    "X11-distribute-modifications-variant",
    "X11-swapped",
    "Xdebug-1.03",
    "Xerox",
    "Xfig",
    "XFree86-1.1",
    "xinetd",
    "xkeyboard-config-Zinoviev",
    "xlock",
    "Xnet",
    "xpp",
    "XSkat",
    "xzoom",
    "YPL-1.0",
    "YPL-1.1",
    "Zed",
    "Zeeff",
    "Zend-2.0",
    "Zimbra-1.3",
    "Zimbra-1.4",
    "Zlib",
    "zlib-acknowledgement",
    "ZPL-1.1",
    "ZPL-2.0",
    "ZPL-2.1",
]

DeprecatedLicenseId = Literal[
    "AGPL-1.0",
    "AGPL-3.0",
    "BSD-2-Clause-FreeBSD",
    "BSD-2-Clause-NetBSD",
    "bzip2-1.0.5",
    "eCos-2.0",
    "GFDL-1.1",
    "GFDL-1.2",
    "GFDL-1.3",
    "GPL-1.0",
    "GPL-1.0+",
    "GPL-2.0",
    "GPL-2.0+",
    "GPL-2.0-with-autoconf-exception",
    "GPL-2.0-with-bison-exception",
    "GPL-2.0-with-classpath-exception",
    "GPL-2.0-with-font-exception",
    "GPL-2.0-with-GCC-exception",
    "GPL-3.0",
    "GPL-3.0+",
    "GPL-3.0-with-autoconf-exception",
    "GPL-3.0-with-GCC-exception",
    "LGPL-2.0",
    "LGPL-2.0+",
    "LGPL-2.1",
    "LGPL-2.1+",
    "LGPL-3.0",
    "LGPL-3.0+",
    "Net-SNMP",
    "Nunit",
    "StandardML-NJ",
    "wxWindows",
]



from __future__ import annotations

import dataclasses
import re
import sys
from dataclasses import dataclass
from functools import wraps
from inspect import isfunction, signature
from typing import (
    Any,
    Callable,
    Dict,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import pydantic
from ruyaml import Optional
from typing_extensions import ParamSpec

if sys.version_info < (3, 10):  # pragma: no cover
    SLOTS: Dict[str, bool] = {}
else:
    SLOTS = {"slots": True}


K = TypeVar("K")
V = TypeVar("V")
NestedDict = Dict[K, "NestedDict[K, V] | V"]

if sys.version_info < (3, 9):  # pragma: no cover
    from functools import lru_cache as cache
    from pathlib import Path

    def files(package_name: str):
        assert package_name == "bioimageio.spec", package_name
        return Path(__file__).parent.parent

else:
    from functools import cache as cache
    from importlib.resources import files as files


def get_format_version_tuple(format_version: Any) -> Optional[Tuple[int, int, int]]:
    if (
        not isinstance(format_version, str)
        or format_version.count(".") != 2
        or any(not v.isdigit() for v in format_version.split("."))
    ):
        return None

    parsed = tuple(map(int, format_version.split(".")))
    assert len(parsed) == 3
    return parsed


def nest_dict(flat_dict: Dict[Tuple[K, ...], V]) -> NestedDict[K, V]:
    res: NestedDict[K, V] = {}
    for k, v in flat_dict.items():
        node: Union[Dict[K, Union[NestedDict[K, V], V]], NestedDict[K, V]] = res
        for kk in k[:-1]:
            if not isinstance(node, dict):
                raise ValueError(f"nesting level collision for flat key {k} at {kk}")
            d: NestedDict[K, V] = {}
            node = node.setdefault(kk, d)  # type: ignore

        if not isinstance(node, dict):
            raise ValueError(f"nesting level collision for flat key {k}")

        node[k[-1]] = v

    return res


FirstK = TypeVar("FirstK")


def nest_dict_with_narrow_first_key(
    flat_dict: Dict[Tuple[K, ...], V], first_k: Type[FirstK]
) -> Dict[FirstK, "NestedDict[K, V] | V"]:
    """convenience function to annotate a special version of a NestedDict.
    Root level keys are of a narrower type than the nested keys. If not a ValueError is raisd.
    """
    nested = nest_dict(flat_dict)
    invalid_first_keys = [k for k in nested if not isinstance(k, first_k)]
    if invalid_first_keys:
        raise ValueError(f"Invalid root level keys: {invalid_first_keys}")

    return nested  # type: ignore


def unindent(text: str, ignore_first_line: bool = False):
    """remove minimum count of spaces at beginning of each line.

    Args:
        text: indented text
        ignore_first_line: allows to correctly unindent doc strings
    """
    first = int(ignore_first_line)
    lines = text.split("\n")
    filled_lines = [line for line in lines[first:] if line]
    if len(filled_lines) < 2:
        return "\n".join(line.strip() for line in lines)

    indent = min(len(line) - len(line.lstrip(" ")) for line in filled_lines)
    return "\n".join(lines[:first] + [line[indent:] for line in lines[first:]])


T = TypeVar("T")
P = ParamSpec("P")


def assert_all_params_set_explicitly(fn: Callable[P, T]) -> Callable[P, T]:
    @wraps(fn)
    def wrapper(*args: P.args, **kwargs: P.kwargs):
        n_args = len(args)
        missing: Set[str] = set()

        for p in signature(fn).parameters.values():
            if p.kind == p.POSITIONAL_ONLY:
                if n_args == 0:
                    missing.add(p.name)
                else:
                    n_args -= 1  # 'use' positional arg
            elif p.kind == p.POSITIONAL_OR_KEYWORD:
                if n_args == 0:
                    if p.name not in kwargs:
                        missing.add(p.name)
                else:
                    n_args -= 1  # 'use' positional arg
            elif p.kind in (p.VAR_POSITIONAL, p.VAR_KEYWORD):
                pass
            elif p.kind == p.KEYWORD_ONLY:
                if p.name not in kwargs:
                    missing.add(p.name)

        assert not missing, f"parameters {missing} of {fn} are not set explicitly"

        return fn(*args, **kwargs)

    return wrapper


def get_os_friendly_file_name(name: str) -> str:
    return re.sub(r"\W+|^(?=\d)", "_", name)


@dataclass
class _PrettyDataClassReprMixin:
    """A mixin that provides a pretty __repr__ for dataclasses

    - leaving out fields that are None
    - leaving out memory locations of functions
    """

    def __repr__(self):
        field_values = {
            f.name: v
            for f in dataclasses.fields(self)
            if (v := getattr(self, f.name)) is not None
        }
        field_str = ", ".join(
            f"{k}=" + (f"<function {v.__name__}>" if isfunction(v) else repr(v))
            for k, v in field_values.items()
        )
        return f"{self.__class__.__name__}({field_str})"


class PrettyPlainSerializer(pydantic.PlainSerializer, _PrettyDataClassReprMixin):
    pass


class PrettyWrapSerializer(pydantic.WrapSerializer, _PrettyDataClassReprMixin):
    pass
